(* Author: Andreas Lochbihler, ETH Zurich
   Author: Florian Haftmann, TU Muenchen 
*)

section \<open>A distributive polynomial representation\<close>

theory Distributive
imports MPoly Monom_Order OAList
begin

text \<open>
  The type @{typ "'a mpoly"} is actually already a distributive representation,
  so we only have to add the monomial order as a representation detail and wrap it in a
  type constructor.

  The implementation will then use associative lists sorted according to the monomial order.
\<close>

type_synonym 'a poly_distr_impl = "(monom \<Rightarrow>\<^sub>0 'a) \<times> monom_order"

typedef (overloaded) 'a poly_distr = "UNIV :: 'a poly_distr_impl set" ..

setup_lifting type_definition_poly_distr

lift_definition Distr :: "'a :: zero poly_distr \<Rightarrow> 'a mpoly"
is "Poly_Mapping.map_key Abs_nat_mapping \<circ> fst :: 'a poly_distr_impl \<Rightarrow> ((nat \<Rightarrow>\<^sub>0 nat) \<Rightarrow>\<^sub>0 'a)" .

instantiation poly_distr :: (monoid_add) plus begin
lift_definition plus_poly_distr :: "'a poly_distr \<Rightarrow> 'a poly_distr \<Rightarrow> 'a poly_distr"
is "\<lambda>(p, mo) (q, mo').
  (Poly_Mapping.map_key Rep_nat_mapping (Poly_Mapping.map_key Abs_nat_mapping p
                                       + Poly_Mapping.map_key Abs_nat_mapping q), mo)" .
instance ..
end

lemma plus_distr: "Distr p + Distr q = Distr (p + q)"
by transfer(clarsimp simp add: Poly_Mapping.map_key_compose o_def Abs_nat_mapping_inverse map_key_id)

lift_definition degree_distr :: "'a :: zero poly_distr \<Rightarrow> nat \<Rightarrow> nat"
  is "\<lambda>(p, mo). degree (MPoly (Poly_Mapping.map_key Abs_nat_mapping p))"
  .

lemma degree_Distr: "degree (Distr p) = degree_distr p"
by(simp add: Distr_def degree_distr_def split_beta)



text \<open>Type of implementation details for distributive representations\<close>

definition implT :: "'a :: zero poly_distr \<Rightarrow> Monom_Order.impl"
where [simp, code del]: "implT _ = Monom_Order.IMPL"

hide_const (open) implT

adhoc_overloading implT Distributive.implT

lemma implT_code [code]:
  "Distributive.implT d = implT (snd (Rep_poly_distr d))"
by simp


text \<open>Logical construction of different flavours of monom producers\<close>

lift_definition monom_distr :: "monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly"
  is "Poly_Mapping.single :: (nat \<Rightarrow>\<^sub>0 nat) \<Rightarrow> _" .

definition monom_distr' :: "monom_order \<Rightarrow> monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly"
where
  [simp, code del]: "monom_distr' _ = monom_distr"
  
lift_definition monom_distr'_ignore :: "(monom_order, monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly) ignore_arg"
  is "monom_distr'"
  by simp

lemma apply_ignore_monom_distr'_ignore [simp]:
  "apply_ignore monom_distr'_ignore = monom_distr'"
  by (fact monom_distr'_ignore.rep_eq)

definition monom_distr'' :: "Monom_Order.impl \<Rightarrow> monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly"
where
  [simp]: "monom_distr'' mo m c = monom_distr m c"

lemma monom_distr''_code [code]:
  "monom_distr'' = switch_monom_order monom_distr'_ignore"
  by transfer (simp add: fun_eq_iff)

definition default_monom_order :: "Monom_Order.impl"
where "default_monom_order = RLEX"

lemma monom_distr_code [code]:
  "monom_distr = monom_distr'' default_monom_order"
by(simp add: monom_distr_def fun_eq_iff)




subsection \<open>Implementation of distributive polynomials\<close>

text \<open>
  Actually, the distributive representation is just a finite map from monomials to coefficients
  parametrised by a comparator with the additional invariant that @{term 0} does not
  occur in the range of the map and the comparator is in fact a monomial order.  Ideally, we
  should factor the map part in a separate type such that it is easy to switch data structures.
  Unfortunately, this will probably cost some efficiency and requires to model all interesting
  ordering relations as data in the code generator.  Therefore, we stick to a monolithic
  implementation with associative lists for now.
\<close>

type_synonym 'a poly_distr_raw = "(monom \<times> 'a) list \<times> monom_order"

lift_definition poly_distr_of_raw :: "'a poly_distr_raw \<Rightarrow> 'a :: zero poly_distr"
is "(\<lambda>(xs, mo). (of_oalist (monom_compare mo) xs, mo))" .

lift_definition raw_of_poly_distr :: "'a :: zero poly_distr \<Rightarrow> 'a poly_distr_raw"
is "(\<lambda>(p, mo). (to_oalist (monom_compare mo) p, mo))" .

definition distr_raw_invar :: "'a :: zero poly_distr_raw \<Rightarrow> bool"
where "distr_raw_invar = (\<lambda>(xs, mo). c.dsorted_by (compare_vimage fst (monom_compare mo)) xs \<and> 0 \<notin> snd ` set xs)"

lemma raw_of_poly_distr [simp]: "distr_raw_invar (raw_of_poly_distr p)"
by transfer(auto simp add: distr_raw_invar_def split_beta dsorted_by_to_oalist zero_notin_to_oalist)

lemma raw_of_poly_distr_inverse [simp, code abstype]:
  "poly_distr_of_raw (raw_of_poly_distr p) = p"
by transfer clarsimp

lemma poly_distr_of_raw_inverse [simp]:
  "distr_raw_invar q \<Longrightarrow> raw_of_poly_distr (poly_distr_of_raw q) = q"
by transfer (auto simp add: distr_raw_invar_def)

lemma poly_distr_rep: "type_definition raw_of_poly_distr poly_distr_of_raw {p. distr_raw_invar p}"
by(unfold_locales)(simp_all add: raw_of_poly_distr raw_of_poly_distr_inverse poly_distr_of_raw_inverse)

interpretation poly_distr_rep: type_definition raw_of_poly_distr poly_distr_of_raw "{p. distr_raw_invar p}"
by(rule poly_distr_rep)

fun plus_distr_raw 
  :: "'a :: {zero, plus} poly_distr_raw \<Rightarrow> 'a poly_distr_raw \<Rightarrow> 'a poly_distr_raw"
where
  "plus_distr_raw (p1, cmp1) (p2, cmp2) =
   (if cmp1 = cmp2 then (oa_zip_with' (monom_compare cmp1) (\<lambda>_. Some) (\<lambda>_. Some)
     (\<lambda>_ x y. let z = x + y in if z = 0 then None else Some z) p1 p2, cmp1)
    else 
     (* if monomials are sorted by different orders, we pick the one of the left operand, 
      but any monomial order would do *)
     let
       cmp = monom_compare cmp1;
       p2' = c.sort_by (compare_vimage fst cmp) p2
     in (oa_zip_with' cmp (\<lambda>_. Some) (\<lambda>_. Some) (\<lambda>_ x y. let z = x + y in if z = 0 then None else Some z) p1 p2', cmp1))"

context
  fixes cmp :: "'a comparator"
  assumes comparator_eq: "comparator_eq cmp"
begin

interpretation c: comparator_eq_on cmp UNIV by(fact comparator_eq)

lemma oa_zip_with'_plus_dest_fun_by:
  fixes plus :: "'a \<Rightarrow> 'b :: monoid_add \<Rightarrow> _"
  defines "plus \<equiv> \<lambda>_ x y. let z = x + y in if z = 0 then None else Some z"
  assumes "finite {x. f x \<noteq> 0}" "finite {x. g x \<noteq> 0}"
  shows "oa_zip_with' cmp (\<lambda>_. Some) (\<lambda>_. Some) plus (dest_fun_by cmp 0 f) (dest_fun_by cmp 0 g) = 
         dest_fun_by cmp 0 (\<lambda>x. f x + g x)"
by(subst oa_zip_with'_dest_fun_by[OF c.comparator_eq_on, where dflt''=0])
  (simp_all add: Let_def assms fun_eq_iff arg_cong[where f="dest_fun_by cmp 0"])

end

lemma plus_distr_raw_to_oalist:
  "plus_distr_raw (to_oalist (monom_compare mo) p, mo)
   (to_oalist (monom_compare mo') q, mo') = (to_oalist (monom_compare mo) (p + q), mo)"
apply(simp add: to_oalist_def sort_by_dest_fun_by oa_zip_with'_plus_dest_fun_by)
apply transfer
apply simp
done

lemma plus_poly_distr_code [code]:
  "raw_of_poly_distr (p + q) = plus_distr_raw (raw_of_poly_distr p) (raw_of_poly_distr q)"
by transfer(clarsimp simp add: map_key_plus map_key_compose o_def Rep_nat_mapping_inverse map_key_id plus_distr_raw_to_oalist simp del: plus_distr_raw.simps)


definition degree_distr_raw :: "'a :: zero poly_distr_raw \<Rightarrow> nat \<Rightarrow> nat"
where
  "degree_distr_raw d v =
    Max (insert 0 (set (map (\<lambda>(m, _). Nat_Mapping.lookup m v) (fst d))))"
-- \<open>
  Maybe a better implementation is possible when we exploit the given monomial ordering.
\<close>

lemma degree_distr_code [code]:
  "degree_distr p = degree_distr_raw (raw_of_poly_distr p)"
  by transfer
    (auto 4 3 simp add: MPoly.degree_def degree_distr_raw_def fun_eq_iff MPoly_inverse keys_map_key vimage_Abs_nat_mapping image_image Nat_Mapping.lookup_def split_def set_to_oalist intro: rev_image_eqI arg_cong)


text \<open>Representation-specific material\<close>

lift_definition monom_distr_impl :: "monom_order \<Rightarrow> monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a poly_distr"
is "\<lambda>(mo :: monom_order) (m :: monom) c. (Poly_Mapping.single m c, mo)" .

lemma monom_distr_impl_code [code]:
  "raw_of_poly_distr (monom_distr_impl mo m c) = (if c = 0 then ([], mo) else ([(m, c)], mo))"
by transfer(simp add: to_oalist_single)

lemma monom_distr'_code [code]:
  "monom_distr' mo m c = Distr (monom_distr_impl mo m c)"
by(cases m)(simp add: monom_distr_impl_def monom_distr_def Distr_def Abs_poly_distr_inverse Abs_nat_mapping_inverse)

subsection \<open>Polynomial multiplication\<close>

(* introduce * on poly_distr *)
instantiation poly_distr :: (semiring_0) times begin
lift_definition times_poly_distr :: "'a poly_distr \<Rightarrow> 'a poly_distr \<Rightarrow> 'a poly_distr"
is "\<lambda>(p, mo) (q, mo').
  (Poly_Mapping.map_key Rep_nat_mapping (Poly_Mapping.map_key Abs_nat_mapping p
                                       * Poly_Mapping.map_key Abs_nat_mapping q), mo)" .
instance 
..
end

(* allow * on mpolys with distributive representation *)
lemma mult_distr: "Distr p * Distr q = Distr (p * q)"
by transfer(clarsimp simp add: Poly_Mapping.map_key_compose o_def Abs_nat_mapping_inverse map_key_id)

(* multiply polynomial with monomial/coefficient pair *)
fun mult_distr_raw_mono
 :: "'a :: {zero, plus, times} poly_distr_raw \<Rightarrow> (monom \<times> 'a) \<Rightarrow> 'a poly_distr_raw"
where
  "mult_distr_raw_mono (p1,cmp1) (m,a) =
     (let
       monos = map (\<lambda>(m0,a0). (m+m0, a*a0)) p1;
       monos0  = c.sort_by (compare_vimage fst (monom_compare cmp1)) monos 
     in (monos0, cmp1))"

(* multiply two polynomials in distributive representation *)
fun mult_distr_raw 
  :: "'a :: {zero, plus, times} poly_distr_raw \<Rightarrow> 'a poly_distr_raw \<Rightarrow> 'a poly_distr_raw"
where
  (* simple definition, but involves repeated sorting of intermediate results *)
  "mult_distr_raw (p1, cmp1) (p2, cmp2) = (* ignore cmp2 *)
    (let ps = map (\<lambda>m. mult_distr_raw_mono (p1, cmp1) m) p2
     in foldr plus_distr_raw  ps ([], cmp1))"

(* auxiliary for lemma mult_distr_raw_to_oalist *)
context
  fixes cmp :: "'a comparator"
  assumes comparator_eq: "comparator_eq cmp"
begin
interpretation c: comparator_eq_on cmp UNIV by(fact comparator_eq)
lemma oa_zip_with'_mult_dest_fun_by:
  fixes mult :: "'a \<Rightarrow> 'b :: {zero,plus,times} \<Rightarrow> _"
  defines "mult \<equiv> \<lambda>_ x y. let z = x + y in if z = 0 then None else Some z"
  assumes "finite {x. f x \<noteq> 0}" "finite {x. g x \<noteq> 0}"
  shows "oa_zip_with' cmp (\<lambda>_. Some) (\<lambda>_. Some) mult (dest_fun_by cmp 0 f) (dest_fun_by cmp 0 g) = 
         dest_fun_by cmp 0 (\<lambda>x. f x * g x)"
sorry
(*
by(subst oa_zip_with'_dest_fun_by[OF c.comparator_eq_on, where dflt''=0])
  (simp_all add: Let_def assms fun_eq_iff arg_cong[where f="dest_fun_by cmp 0"])
*)
end

(* auxiliary for code generation lemma *)
lemma mult_distr_raw_to_oalist:
  "mult_distr_raw (to_oalist (monom_compare mo) p, mo)
   (to_oalist (monom_compare mo') q, mo') = (to_oalist (monom_compare mo) (p * q), mo)"
apply(simp add: to_oalist_def sort_by_dest_fun_by oa_zip_with'_mult_dest_fun_by)
apply transfer
sorry
(*
apply(simp add: to_oalist_def sort_by_dest_fun_by oa_zip_with'_plus_dest_fun_by)
apply transfer
apply simp
done
*)

(* auxiliary for code generation lemma *)
context
  fixes f :: "'a \<Rightarrow> 'b"
  assumes inj_f [transfer_rule]: "inj f"
begin
lemma map_key_mult:
  "map_key f (p * q) = map_key f p * map_key f q"
(* by transfer (simp add: fun_eq_iff) *)
sorry
end

lemma mult_poly_distr_code [code]:
  "raw_of_poly_distr (p * q) = mult_distr_raw (raw_of_poly_distr p) (raw_of_poly_distr q)"
by transfer(clarsimp simp add: map_key_mult map_key_compose o_def Rep_nat_mapping_inverse 
     map_key_id mult_distr_raw_to_oalist simp del: mult_distr_raw.simps)

subsection \<open>Long division of polynomials\<close>

subsection \<open>Conversion functions\<close>

lift_definition monom_order_of_distr :: "'a :: zero poly_distr \<Rightarrow> monom_order"
is "snd :: 'a poly_distr_impl \<Rightarrow> _" .

lemma monom_order_of_distr_code [code]:
  "monom_order_of_distr p = snd (raw_of_poly_distr p)"
by transfer clarsimp

lift_definition convert_monom_order :: "monom_order \<Rightarrow> 'a poly_distr \<Rightarrow> 'a :: zero poly_distr"
is "\<lambda>mo :: monom_order. \<lambda>(p, _ :: monom_order). (p, mo)" .

lemma convert_monom_order_code [code]:
  "raw_of_poly_distr (convert_monom_order mo p) = (c.sort_by (compare_vimage fst (monom_compare mo)) (fst (raw_of_poly_distr p)), mo)"
apply(transfer fixing: mo)
apply clarsimp
apply(simp add: to_oalist_def dest_fun_by_def)
sorry


definition distr_of :: "monom_order \<Rightarrow> 'a :: zero mpoly \<Rightarrow> 'a poly_distr"
where "distr_of mo p = Abs_poly_distr (Poly_Mapping.map_key Rep_nat_mapping (MPoly.mapping_of p), mo)"

lemma distr_of_Distr:
  "distr_of mo (Distr p) = (if mo = monom_order_of_distr p then p else convert_monom_order mo p)"
by(simp add: distr_of_def Distr_def MPoly_inverse map_key_compose o_def Rep_nat_mapping_inverse map_key_id monom_order_of_distr_def Rep_poly_distr_inverse convert_monom_order_def split_beta)

lemma Distr_distr_of: "Distr (distr_of mo p) = p"
by(simp add: distr_of_def Distr_def mapping_of_inverse map_key_compose o_def Abs_nat_mapping_inverse map_key_id monom_order_of_distr_def Abs_poly_distr_inverse convert_monom_order_def split_beta)

subsection \<open>Old stuff\<close>

definition lookup0 :: "('a \<times> 'b::zero) list \<Rightarrow> 'a \<Rightarrow> 'b"
where
  "lookup0 xs k = (case map_of xs k of None \<Rightarrow> 0 | Some v \<Rightarrow> v)"

definition zero_bound :: "(nat \<Rightarrow> nat) \<Rightarrow> nat"
where
  "zero_bound m = Min {n. \<forall>q\<ge>n. m q = 0}"

(*lemma
  "zero_bound (lookup0 [])*)

primrec make :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> ('a \<times> 'b) list"
where
  "make f [] = []"
| "make f (k # ks) = (k, f k) # make f ks"

definition monom_function_of :: "(nat \<Rightarrow> nat) \<Rightarrow> (nat \<times> nat) list"
where
  "monom_function_of m = make m [0..<zero_bound m]"

(*lemma
  "distinct xs \<Longrightarrow> monom_function_of (lookup0 xs) = xs"
  apply (simp add: monom_function_of_def)
  apply (induct xs)
  apply simp_all*)                                               

(*broken at end of Florian's session
type_synonym 'a poly = "((nat, nat) alist, 'a) alist"

lift_definition MPoly :: "'a poly \<Rightarrow> 'a mpoly"
  is "\<lambda>xs m. lookup0 xs"

hide_type (open) poly monom MPoly
*)

end