(* Author: Andreas Lochbihler, ETH Zurich
   Author: Florian Haftmann, TU Muenchen 
   Author: Walther Neuper, TU Graz
*)
theory Examples
imports Complex_Main MPoly_Impl Rings
begin

section \<open>An abstract @{typ "'a mpoly"}\<close>
text \<open>
  Polynomials are, as usual in mathematics, defined as a mapping on naturals
  which has only finitely many values unequal zero:

  @{text "typedef 'a \<Rightarrow>\<^sub>0 'b = {f :: 'a \<Rightarrow> 'b::zero. finite {x. f x \<noteq> 0}}"}
  @{text "typedef 'a mpoly = UNIV :: ((nat \<Rightarrow>\<^sub>0 nat) \<Rightarrow>\<^sub>0 'a::zero) set"}

  These polynomials are called "abstract", because they primarily serve definitions
  close to standard textbooks, to formulate algorithms as abstract as possible and 
  to have concise and readable proofs.
\<close>

subsection \<open>Definitions and theories close to standard textbooks\<close>
text \<open>
  Isabelle's language for definitions, proofs etc resembles traditional textbooks.
  For instance, this is the definition for "primitive polynomials":
\<close>
definition primitive :: "'a::{ring_div,GCD.Gcd} mpoly \<Rightarrow> bool"
where "primitive p \<longleftrightarrow> Gcd (coeffs p) = 1"

subsection \<open>Algebraic operations on @{typ "'a mpoly"}\<close>
text \<open>
  Algorithms, even abstractly formulated ones, involve algebraic operations.
  @{typ "'a mpoly"} shall be proven to form commutative ring with one and 
  an integral domain with linear order.

  Isabelle provides means to associate operations with traditional notation.
  For instance, this is an algorithm, which uses the already present "+":
\<close>
definition double :: "'a::monoid_add mpoly \<Rightarrow> 'a mpoly" (*TODO ring*)
where "double p = p + p"

definition square :: "'a::semiring_0 mpoly \<Rightarrow> 'a mpoly"
where "square p = p * p"

subsection \<open>Concise and readable proofs\<close>
text \<open>
  This is a simple theorem with proof for demonstration purposes:
\<close>
lemma double_not_primitive:
  assumes "q = double p"
  shows "\<not> primitive q"
proof
  assume "primitive q"
  then have *: "Gcd (coeffs q) = 1" by (simp add: primitive_def)
  from assms have "coeffs q = coeffs (double p)" by simp
  then have "coeffs q = coeffs (p + p)" by (simp add: double_def)
  then have "coeffs q = Groups.times 2 ` coeffs p" sorry
  then have "Gcd (coeffs q) = 2 * Gcd (coeffs p)" sorry
  then have "Gcd (coeffs q) \<noteq> 1" sorry
  with * show False by simp
qed


subsection \<open>Executable operations for technical reasons\<close>
text \<open>
  Some of Isabelle's tools rely on executability of definitions, for instance
  the counter example generator "Quickcheck".
  So definitions are executable, which is useful for prototyping, too. 
  For instance:
\<close>
value
  "let p = M' (DISTR RLEX) [2, 1] (3 / 5::rat) + M' (DISTR RLEX) [0, 3] (7 / 3) + M' (DISTR RLEX) [1, 1] 1
  in (double p)"

value "let p = M [2, 1] (3 / 5::rat) + M [0, 3] (7 / 3) + M [1, 1] 1
  in (double p)"

section \<open>Several polynomial representations\<close>
text \<open>
  The planned structure of the polynomial package anticipates several different
  polynomial representations and several different monomial orders.
\<close>
subsection \<open>Distributive and recursive representations\<close>
text \<open>
  Presently there are the types @{typ "'a poly_distr"} and @{typ "'a poly_rec"},
  a distributive and a recursive representation respectively:

  typedef 'a poly_distr = "UNIV :: (nat nat_mapping, 'a) poly_mapping set" ..
  typedef 'a poly_rec = "{p :: 'a :: zero raw_poly_rec. no_trailing_zeros_raw_poly_rec p}"

  @{typ "'a poly_distr"} is defined directly by @{typ "'a nat_mapping"},
  which is derived from @{typ "'a \<Rightarrow>\<^sub>0 'b"}, which in turn serves
  derivation of the abstract @{typ "'a mpoly"}.
  @{typ "'a poly_rec"} cannot (yet?) be implemented by Isabelles standard
  mechanisms and requires more tedious coding.

  @{typ "'a nat_mapping"} introduces "dense representation" and 
  "sparse representation" in a general way, which can be re-used for 
  further refined representations.
\<close>

subsection \<open>Monomial orders\<close>
text \<open>
  Certain algorithms, for instance the Groebner Bases algorithm, require
  distributive representation with a variety of monomial orders for
  efficient implementations.

  So distributive representations all come along with a monomial order:

  type_synonym 'a poly_distr_raw = "(monom \<times> 'a) list \<times> monom_order"

  Presently there is one monomial order, the lexicographic order.
\<close>

subsection \<open>Abstract algorithm called with different representations\<close>
text \<open>
  Although the logical setup for the representations is highly elaborated
  and challenges the limits of what can be done in Isabelle presently,
  developers of algorithms are not exposed to these intricacies.

  Rather, developers use simple and concise code as shown in the definition
  of @{const double} above. Associating specific representations with
  algorithms is simple, too. This is shown in the examples below.
  which first defines somes polynomials @{text p011} .. @{text p013}
  (using a preliminary syntax):
\<close>
definition p011 :: "int mpoly" --\<open>distributive sparse\<close>
where "p011 = monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 0 1) 2 +
              monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 2 4) 3"

definition p012 :: "int mpoly" --\<open>distributive dense\<close>
where "p012 = monom' (DISTR RLEX) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (DISTR RLEX) (Nat_Mapping.single' DENSE 2 4) 3"

definition p013 :: "int mpoly" --\<open>recursive dense\<close>
where "p013 = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (REC DENSE) (Nat_Mapping.single' DENSE 2 4) 3"

definition p014 :: "int mpoly" --\<open>recursive sparse\<close>
where "p014 = monom' (REC SPARSE) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (REC SPARSE) (Nat_Mapping.single' DENSE 2 4) 3"

subsubsection \<open>Addition\<close>
value p011
export_code  p011 in SML module_name P011

value "double p011" --\<open>result is distributive sparse\<close>
value "double p012" --\<open>result is distributive dense\<close>
value "double p013" --\<open>result is recursive dense\<close>  (*ok*)
value "double p014" --\<open>result is recursive sparse\<close> (*ok*)

(*value "p011 + p013" TODO LATER: conversions*)

subsubsection \<open>Multiplication\<close>

value "smult 2 p013"
(*= "Rec (Powers (Dense (Abs_nat_mapping_dense [
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 6]))])), 
  Coeff 4])))"*)
value "smult 0 p013"
(*= "Rec (Coeff 0)"*)
value "smult 2 p014"
(*"Rec (Powers (Sparse (Abs_nat_mapping_sparse [
  (0, Powers (Sparse (Abs_nat_mapping_sparse [
    (0, Powers (Sparse (Abs_nat_mapping_sparse [(Suc (Suc (Suc (Suc 0))), Coeff 6)])))]))), 
  (Suc 0, Coeff 4)])))"*)
lemma "smult 2 p013 = smult 2 p014"
by(simp_all add: p013_def p014_def)

definition p123 :: "int mpoly" --\<open>recursive dense\<close>
where "p123 = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 0) 1 +
              monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 2) 3"
definition p45 :: "int mpoly" --\<open>recursive dense\<close>
where "p45  = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 0) 4 +
              monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 5"
value "p123 * p45" 
  --\<open>= "Rec (Powers (Dense (Abs_nat_mapping_dense [Coeff 4, Coeff 13, Coeff 22, Coeff 15])))" OK\<close>

definition p2x3y :: "int mpoly" --\<open>recursive dense\<close>
where "p2x3y = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 2 +
               monom' (REC DENSE) (Nat_Mapping.single' DENSE 1 1) 3"
definition p4x5y :: "int mpoly" --\<open>recursive dense\<close>
where "p4x5y = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 4 +
               monom' (REC DENSE) (Nat_Mapping.single' DENSE 1 1) 5"
(* debugging functions:
declare [[code drop: inc_power_rec]]
value "p2x3y * p4x5y" 
(*"Rec (inc_power_rec DENSE (Suc 0)
 (Powers
   (Dense (Abs_nat_mapping_dense [Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 15])), Coeff 12])) +
  inc_power_rec DENSE (Suc 0) (Coeff 0)) +
inc_power_rec DENSE 0
 (inc_power_rec DENSE 0
   (Powers
     (Dense (Abs_nat_mapping_dense [Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 10])), Coeff 8])) +
    inc_power_rec DENSE 0 (Coeff 0))))" 
with ERROR from f482e1c2f6ec: 
(Powers [Powers [Coeff 0, Coeff 2], Coeff 3]) * (Coeff Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) 
\<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10], Powers [Coeff 0, Coeff 12], Coeff 15]) 
CORRECT would be
\<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15]) 
 *)
*)
value "p2x3y * p4x5y" --\<open>
"Rec (                                        (* c449292e2133 NOT OK 1st result with foldr_step  *)
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 0, Coeff 15])), 
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 12])),
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 10])), 
    Coeff 8])))"                                
"Rec (                                        (* 381393f4c823 NOT OK result with foldr *)
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 15])), 
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 12, Coeff 10])), 
    Coeff 8])))"
"Rec (                                        (* correct result *)
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 0, Coeff 15])), 
    Powers (Dense (Abs_nat_mapping_dense [
      Coeff 0, Coeff 22])), 
    Coeff 8])))"\<close>

(* multiplication: distributive dense *)
value "p011"
value "square p011"

(* multiplication: distributive sparse *)
value "p012"
value "square p012"

(* multiplication: generated code *)
export_code  square in SML module_name Square

subsection \<open>Conversions\<close>

value "distr_cast SPARSE rlex_mo p014"
value "distr_cast DENSE rlex_mo p013"
value "distr_cast SPARSE rlex_mo p012"
value "distr_cast SPARSE lex_mo p011"

text \<open>Eliminate uninteresting morphisms in printing of value\<close>

definition [code del]: "Distr_impl_pretty = Distr"
definition [code del]: "poly_distr_of_raw_pretty = poly_distr_of_raw"
definition [code del]: "Abs_nat_mapping_sparse_pretty = Abs_nat_mapping_sparse"
definition [code del]: "Abs_nat_mapping_dense_pretty = Abs_nat_mapping_dense"
notation (output)
  Distr_impl_pretty ("_") and
  poly_distr_of_raw_pretty ("_") and
  Abs_nat_mapping_sparse_pretty ("_") and
  Abs_nat_mapping_dense_pretty ("_")

lemmas mpoly_pretty[symmetric] =
  Distr_impl_pretty_def
  poly_distr_of_raw_pretty_def
  Abs_nat_mapping_sparse_pretty_def
  Abs_nat_mapping_dense_pretty_def

declare mpoly_pretty[code_post]

value "double p011" (*TODO LATER: improve pretty printing after finish of poly pack*)


subsection \<open>Verification of algorithms\<close>
text \<open>
  Verification of algorithms (together with generation of efficient code
  within the same system, see below) is considered the distinguishing feature 
  of the planned polynomial package.

  The type of polynomials intended for concise proofs is the abstract @{typ "'a mpoly"}.
  The common source @{typ "'a \<Rightarrow>\<^sub>0 'b"} for deriving both, 
  @{typ "'a mpoly"} and all representations like @{typ "'a poly_distr"}
  and  @{typ "'a poly_rec"} simplifies transfer of validity to representations.

  lemma double_not_primitive is an example: The respective proof
  used @{typ "'a mpoly"}, and the validity transfers to @{typ "'a poly_distr"}
  and to @{typ "'a poly_rec"} by automatically provable lemmas:
\<close>
lemma "p011 = p012" "p012 = p013" "p013 = p014"
by(simp_all add: p011_def p012_def p013_def p014_def)

text \<open>
  Of course, this will be proven for all members of the type; then this proves logical 
  equivalence of all representations follows and thus any theorems transfer.
\<close>

section \<open>Generation of efficient code for specific purposes\<close>
text \<open>
  Generation of efficient code (and verification within the same system, see above)
  is considered the distinguishing feature of the planned polynomial package.
\<close>

subsection \<open>Code from an abstract algorihm\<close>
text \<open>
  The function @{const double} is abstract in the sense, that it
  does not care about polynomial representation at all. Thus the 
  generated code associates specific representations with respective 
  operations automatically at run time:
\<close>
export_code double in SML module_name Double

subsection \<open>Specific code for specific coefficient domains\<close>
text \<open>
  The algorithms shown so far abstract over coefficient domains by
  use of the type variable @{typ 'a}. This variable can be instantiated
  with types fulfilling the ring axioms:
\<close>
definition double_int :: "int mpoly \<Rightarrow> int mpoly" --\<open>integers\<close>
where "double_int p = p + p"
export_code double_int p011 in SML module_name Double_Int

definition double_rat :: "rat mpoly \<Rightarrow> rat mpoly" --\<open>rational numbers\<close>
where "double_rat p = p + p"
export_code double_rat in SML module_name Double_Rat

definition double_complex :: "complex mpoly \<Rightarrow> complex mpoly" --\<open>complex numbers\<close>
where "double_complex p = p + p"
export_code double_complex in SML module_name Double_Complex

subsection \<open>Algorithms bound to specific representations\<close>
text \<open>
  Certain algorithms are known to perform efficiently only on a certain
  representation.
  If this would be the case for the above @{const double}, then we could
  define more specifically:
\<close>
definition double_distr :: "int poly_distr \<Rightarrow> int poly_distr"
where "double_distr p = p + p"
export_code double_distr in SML module_name Double_Distr

text \<open>
  The code could even be "double_distr p = p + p" as soon as poly_distr
  itself has been instantiated to a ring. TODO?

  Or given other requirements for the algorithm we could define: 
\<close>
definition double_rec :: "int poly_rec \<Rightarrow> int poly_rec"
where "double_rec p = p + p"
export_code double_rec in SML module_name Double_Rec

end
