(* Author: Andreas Lochbihler, ETH Zurich *)

theory MPoly_Impl imports
  Distributive
  Recursive
begin

code_datatype Distr Rec

lemmas [code del] =
  total_degree.abs_eq
  zero_mpoly.abs_eq
  one_mpoly.abs_eq
  monom.abs_eq
  plus_mpoly.abs_eq
  uminus_mpoly.abs_eq
  minus_mpoly.abs_eq
  smult.abs_eq
  times_mpoly.abs_eq
  sdiv.abs_eq
  coeffs.abs_eq

lemmas [code] =
  plus_distr plus_rec smult_rec

datatype impl = IMPL

lemma impl_eq [no_atp]: "x = IMPL" by(cases x)(simp)

text \<open>
  Simplification procedure for @{thm [source] impl_eq}.  Cannot use
  this rule directly --- it loops!
\<close>

simproc_setup impl_eq ("x::impl") = \<open>fn _ => fn _ => fn ct =>
  case Thm.term_of ct of Const (@{const_name IMPL}, _) => NONE | _ => SOME (mk_meta_eq @{thm impl_eq})
\<close>

lemma equal_impl_code [code]:
  "HOL.equal (x :: impl) y = True"
by(simp add: equal_eq)

definition REC :: "Nat_Mapping.impl \<Rightarrow> impl" where [simp, code del]: "REC _ = IMPL"
definition DISTR :: "Monom_Order.impl \<Rightarrow> impl" where [simp, code del]: "DISTR _ = IMPL"

code_datatype REC DISTR

definition rec_impl_at :: "Nat_Mapping.impl \<Rightarrow> 'a :: zero poly_rec \<Rightarrow> Nat_Mapping.impl"
where [simp, code del]: "rec_impl_at _ _ = Nat_Mapping.IMPL"

lemma rec_impl_at_code [code]:
  "rec_impl_at dflt (Coeff c) = dflt"
  "rec_impl_at dflt (Powers ps) = implT ps"
by simp_all

definition implT :: "'a mpoly \<Rightarrow> impl" where [simp, code del]: "implT _ = IMPL"

hide_const (open) implT

adhoc_overloading implT MPoly_Impl.implT


lemma implT_code [code]:
  "implT (Rec r) = REC (rec_impl_at SPARSE r)"
  "implT (Distr d) = DISTR (implT d)"
by simp_all

hide_const (open) IMPL
hide_type (open) impl


lift_definition monom' :: "MPoly_Impl.impl \<Rightarrow> monom \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly"
is "\<lambda>_. monom" .

lemma monom'_REC:
  "monom' (REC impl) m c = Rec (monom_rec' impl m c)"
sorry

lemma monom'_DISTR:
  "monom' (DISTR mo) m c = monom_distr'' mo m c"
  -- \<open>If we define @{const monom_distr''} in @{theory Distributive},
        we might also want a corresponding version @{text monom_rec} in @{theory Recursive}\<close>
sorry

lemmas [code] = monom'_REC monom'_DISTR


definition M' :: "MPoly_Impl.impl \<Rightarrow> nat list \<Rightarrow> 'a :: zero \<Rightarrow> 'a mpoly"
where [simp, code del]: "M' _ = M"

lemma M'_code [code]: "M' impl xs = monom' impl (Nat_Mapping.of_list DENSE xs)"
unfolding M'_def Nat_Mapping.of_list_def by transfer(simp add: M_def)


subsection \<open>Conversion functions between representations\<close>

context 
  fixes mo :: monom_order
begin

function distr_of_rec_aux :: "nat \<Rightarrow> monom \<Rightarrow> 'a :: monoid_add poly_rec \<Rightarrow> 'a poly_distr"
where
  "distr_of_rec_aux n m (Coeff c) = monom_distr_impl mo m c"
| "distr_of_rec_aux n m (Powers ps) = Nat_Mapping_Recursive.foldr 
     (\<lambda>exp p rest. rest + (distr_of_rec_aux (Suc n) (Nat_Mapping.update n exp m) p))
     ps (monom_distr_impl mo 0 0)"
by(case_tac "snd (snd x)") fastforce+
termination by lexicographic_order

definition distr_of_rec :: "'a :: monoid_add poly_rec \<Rightarrow> 'a poly_distr"
where "distr_of_rec = distr_of_rec_aux 0 0"

end

definition distr_of_rec_aux' :: "Nat_Mapping.impl \<Rightarrow> monom_order \<Rightarrow> nat \<Rightarrow> monom \<Rightarrow> 'a :: monoid_add poly_rec \<Rightarrow> 'a poly_distr"
where [simp, code del]: "distr_of_rec_aux' _ = distr_of_rec_aux"

lemma distr_of_rec_aux'_code [code]:
  "distr_of_rec_aux' ds mo n m (Coeff c) = monom_distr_impl mo m c"
  "distr_of_rec_aux' ds mo n m (Powers ps) = Nat_Mapping_Recursive.foldr 
     (\<lambda>exp p rest. rest + (distr_of_rec_aux' ds mo (Suc n) (Nat_Mapping.update n exp m) p))
     ps (monom_distr_impl mo (Nat_Mapping.of_list ds []) 0)"
  by simp_all

definition distr_of_rec' :: "Nat_Mapping.impl \<Rightarrow> monom_order \<Rightarrow> 'a :: monoid_add poly_rec \<Rightarrow> 'a poly_distr"
where [simp, code del]: "distr_of_rec' _ = distr_of_rec"

lemma distr_of_rec'_code [code]:
  "distr_of_rec' ds mo = distr_of_rec_aux' ds mo 0 (Nat_Mapping.of_list ds [])"
  by (simp add: distr_of_rec_def)

lemma distr_of_Rec:
  "distr_of mo (Rec p) = distr_of_rec mo p"
sorry

definition distr_of' :: "Nat_Mapping.impl \<Rightarrow> monom_order \<Rightarrow> 'a :: zero mpoly \<Rightarrow> 'a poly_distr"
where [simp, code del]: "distr_of' _ = distr_of"

lemma distr_of'_code [code]:
  -- \<open>Specification of monomial representation (dense/sparse) is ignored if we already have a distributive representation\<close>
  "distr_of' ds mo (Distr p) = (if mo = monom_order_of_distr p then p else convert_monom_order mo p)"
  "distr_of' ds mo (Rec q) = distr_of_rec' ds mo q"
by(simp_all add: distr_of_Distr distr_of_Rec)



definition distr_cast :: "Nat_Mapping.impl \<Rightarrow> monom_order \<Rightarrow> 'a mpoly \<Rightarrow> 'a mpoly"
where [simp]: "distr_cast _ _ = (\<lambda>p. p)"

lemma distr_cast_code [code]: "distr_cast ds mo p = Distr (distr_of' ds mo p)"
by(simp add: Distr_distr_of)


subsection \<open>Input and output representations for generated code\<close>

type_synonym 'a mpoly_exchange = "(integer list * 'a) list"

value "[([0, 1], 2), ([2, 4], 3)]::int mpoly_exchange" (*see Examples.thy*)

definition to_mpoly :: "MPoly_Impl.impl => 'a mpoly_exchange => 'a::zero mpoly"
where "to_mpoly ds p = 0"

definition from_mpoly :: "MPoly_Impl.impl => 'a::zero mpoly => 'a mpoly_exchange"
where "from_mpoly ds p = []"

value "xxx" --\<open>\<close>
value "xxx" --\<open>\<close>
value "xxx" --\<open>\<close>

end