(* Author: Andreas Lochbihler, ETH Zurich
   Author: Florian Haftmann, TU Muenchen 
*)

section \<open>Nat mapping: specialisation of polynomial mappings to key nat\<close>

theory Nat_Mapping 
imports 
  OAList_Poly_Mapping
  "~~/src/HOL/Library/Product_Lexorder"
  "~~/src/HOL/Library/List_lexord"
  MPoly_ImplT
begin

subsection \<open>Type definition\<close>

typedef (overloaded) 'a nat_mapping = "UNIV :: (nat \<Rightarrow>\<^sub>0 'a) set" ..

type_notation nat_mapping ("(\<nat> \<Rightarrow>\<^sub>0 /_)" [0] 0)

setup_lifting type_definition_nat_mapping


subsection \<open>Additive structure\<close>
  
instantiation nat_mapping :: (zero) zero
begin

lift_definition zero_nat_mapping :: "'a nat_mapping"
  is "0 :: nat \<Rightarrow>\<^sub>0 'a" .

instance ..

end

instantiation nat_mapping :: (monoid_add) monoid_add
begin

lift_definition plus_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> 'a nat_mapping"
  is "Groups.plus :: (nat \<Rightarrow>\<^sub>0 'a) \<Rightarrow> _" .

instance
  by intro_classes (transfer, simp add: ac_simps)+

end

instance nat_mapping :: (comm_monoid_add) comm_monoid_add
  by intro_classes (transfer, simp add: fun_eq_iff ac_simps)+

instantiation nat_mapping :: (cancel_comm_monoid_add) cancel_comm_monoid_add
begin

lift_definition minus_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> 'a nat_mapping"
  is "Groups.minus :: (nat \<Rightarrow>\<^sub>0 'a) \<Rightarrow> _" .

instance
  by intro_classes (transfer, simp add: fun_eq_iff diff_diff_add)+

end
  
instantiation nat_mapping :: (ab_group_add) ab_group_add
begin

lift_definition uminus_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping"
  is "Groups.uminus :: (nat \<Rightarrow>\<^sub>0 'a) \<Rightarrow> _" .

instance
  by intro_classes (transfer, simp add: fun_eq_iff ac_simps)+

end


subsection \<open>Multiplicative structure\<close>

instantiation nat_mapping :: (zero_neq_one) zero_neq_one
begin

lift_definition one_nat_mapping :: "'a nat_mapping"
  is "1 :: nat \<Rightarrow>\<^sub>0 'a" .

instance
  by intro_classes (transfer, simp add: fun_eq_iff)

end

instantiation nat_mapping :: (semiring_0) semiring_0
begin

lift_definition times_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> 'a nat_mapping"
  is "Groups.times :: (nat \<Rightarrow>\<^sub>0 'a) \<Rightarrow> _" .

instance
  by intro_classes (transfer, simp add: algebra_simps)+

end

instance nat_mapping :: (comm_semiring_0) comm_semiring_0
  by intro_classes (transfer, simp add: algebra_simps)+

instance nat_mapping :: (semiring_0_cancel) semiring_0_cancel
  ..

instance nat_mapping :: (comm_semiring_0_cancel) comm_semiring_0_cancel
  ..

instance nat_mapping :: (semiring_1) semiring_1
  by intro_classes (transfer, simp)+

instance nat_mapping :: (comm_semiring_1) comm_semiring_1
  by intro_classes (transfer, simp)

instance nat_mapping :: (semiring_1_cancel) semiring_1_cancel
  ..

(*instance nat_mapping :: (comm_semiring_1_cancel) comm_semiring_1_cancel
  ..*)

instance nat_mapping :: (ring) ring
  ..

instance nat_mapping :: (comm_ring) comm_ring
  ..

instance nat_mapping :: (ring_1) ring_1
  ..

instance nat_mapping :: (comm_ring_1) comm_ring_1
  ..

  
subsection \<open>Single-point mappings\<close>  

lift_definition single :: "nat \<Rightarrow> 'a \<Rightarrow> 'a::zero nat_mapping"
  is Poly_Mapping.single .

lemma inj_single [iff]:
  "inj (single k)"
proof (rule injI, transfer)
  fix a b :: 'a and k :: nat
  assume "Poly_Mapping.single k a = Poly_Mapping.single k b"
  with injD [of "Poly_Mapping.single k" a b]
  show "a = b" by simp
qed

lemma single_zero [simp]:
  "single k 0 = 0"
  by transfer simp

lemma single_one [simp]:
  "single 0 1 = 1"
  by transfer simp

lemma single_add:
  "single k (a + b) = single k a + single k b"
  by transfer (simp add: single_add)

lemma single_uminus:
  "single k (- a) = - single k a"
  by transfer (simp add: single_uminus)

lemma single_diff:
  "single k (a - b) = single k a - single k b"
  by transfer (simp add: single_diff)

lemma single_numeral [simp]:
  "single 0 (numeral n) = numeral n"
  by (induct n) (simp_all only: numeral.simps numeral_add single_zero single_one single_add)

lemma single_of_nat [simp]:
  "single 0 (of_nat n) = of_nat n"
  by (induct n) (simp_all add: single_add)

lemma of_nat_single:
  "of_nat = single 0 \<circ> of_nat"
  by (simp add: fun_eq_iff)

lemma mult_single:
  "single k a * single l b = single (k + l) (a * b)"
  by transfer (simp add: mult_single)

instance nat_mapping :: (semiring_char_0) semiring_char_0
  by intro_classes (auto intro: inj_comp inj_of_nat simp add: of_nat_single)

instance nat_mapping :: (ring_char_0) ring_char_0
  ..

lemma single_of_int [simp]:
  "single 0 (of_int k) = of_int k"
  by (cases k) (simp_all add: single_diff single_uminus)

  
subsection \<open>Integral domains\<close>

instance nat_mapping :: (ring_no_zero_divisors) ring_no_zero_divisors
  by intro_classes (transfer, simp)

instance nat_mapping :: (ring_1_no_zero_divisors) ring_1_no_zero_divisors
  ..

instance nat_mapping :: (idom) idom
  ..


subsection \<open>Mapping order\<close>

instantiation nat_mapping :: ("{linorder, zero}") linorder
begin

lift_definition less_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> bool"
  is less .

lift_definition less_eq_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> bool"
  is less_eq .

instance
  by intro_classes (transfer, (rule less_le_not_le linear order_refl|erule (1) order_trans order_antisym))+

end

instance nat_mapping :: ("{ordered_comm_monoid_add, ordered_ab_semigroup_add_imp_le, linorder}") ordered_ab_semigroup_add
  by intro_classes (transfer, fact add_left_mono)

instance nat_mapping :: ("{ordered_comm_monoid_add, ordered_ab_semigroup_add_imp_le, cancel_comm_monoid_add, linorder}") linordered_cancel_ab_semigroup_add
  ..

instance nat_mapping :: ("{ordered_comm_monoid_add, ordered_ab_semigroup_add_imp_le, cancel_comm_monoid_add, linorder}") ordered_comm_monoid_add
  ..


subsection \<open>Fundamental mapping notions\<close>

lift_definition lookup :: "'a :: zero nat_mapping \<Rightarrow> nat \<Rightarrow> 'a"
  is "Poly_Mapping.lookup" .

lift_definition keys :: "'a :: zero nat_mapping \<Rightarrow> nat set"
  is "Poly_Mapping.keys" .

lift_definition range :: "'a :: zero nat_mapping \<Rightarrow> 'a set"
  is "Poly_Mapping.range" .

lemma finite_keys [simp]:
  "finite (keys f)"
  by transfer simp

lemma not_in_keys_iff_lookup_eq_zero [simp]:
  "k \<notin> keys f \<longleftrightarrow> lookup f k = 0"
  by transfer simp

lemma lookup_not_eq_zero_eq_in_keys [simp]:
  "lookup f k \<noteq> 0 \<longleftrightarrow> k \<in> keys f"
  by transfer simp

lemma lookup_eq_zero_in_keys_contradict [dest]:
  "lookup f k = 0 \<Longrightarrow> \<not> k \<in> keys f"
  by simp

lemma in_keys_lookup_in_range [simp]:
  "k \<in> keys f \<Longrightarrow> lookup f k \<in> range f"
  by transfer simp

lemma lookup_zero [simp]:
  "lookup 0 n = 0"
  by transfer simp

lemma keys_zero [simp]:
  "keys 0 = {}"
  by transfer simp

lemma range_zero [simp]:
  "range 0 = {}"
  by transfer simp

lemma lookup_add:
  "lookup (f + g) n = lookup f n + lookup g n"
  by transfer (fact lookup_add)

lemma keys_add_subset:
  "keys (f + g) \<subseteq> keys f \<union> keys g"
  by transfer (fact keys_add_subset)

lemma lookup_one:
  "lookup 1 n = (1 when n = 0)"
  by transfer (fact lookup_one)

lemma lookup_one_zero [simp]:
  "lookup 1 0 = 1"
  by (simp add: lookup_one)

lemma keys_one [simp]:
  "keys 1 = {0}"
  by transfer simp

lemma range_one [simp]:
  "range 1 = {1}"
  by transfer simp

lemma lookup_single [simp]:
  "lookup (single k v) n = (v when k = n)"
  by transfer (fact lookup_single)

lemma keys_single [simp]:
  "keys (single k v) = (if v = 0 then {} else {k})"
  by transfer simp

lemma range_single [simp]:
  "range (single k v) = (if v = 0 then {} else {v})"
  by transfer (auto simp add: when_def)

lemma lookup_mult:
  "lookup (f * g) k = (\<Sum>l. lookup f l * (\<Sum>q. lookup g q when k = l + q))"
  by transfer (fact lookup_mult)

lemma keys_mult:
  "keys (f * g) \<subseteq> {a + b | a b. a \<in> keys f \<and> b \<in> keys g}"
  by transfer (fact keys_mult)


subsection \<open>Degree\<close>

lift_definition degree :: "(\<nat> \<Rightarrow>\<^sub>0 'a::zero) \<Rightarrow> nat"
  is Poly_Mapping.degree .

lemma degree_zero [simp]:
  "degree 0 = 0"
  by transfer simp

lemma degree_one [simp]:
  "degree 1 = 1"
  by transfer simp

lemma degree_single_zero [simp]:
  "degree (single k 0) = 0"
  by transfer simp

lemma degree_single_not_zero [simp]:
  "v \<noteq> 0 \<Longrightarrow> degree (single k v) = Suc k"
  by transfer simp

lemma degree_zero_iff [simp]:
  "degree f = 0 \<longleftrightarrow> f = 0"
  by transfer simp

lemma degree_greater_zero_in_keys:
  assumes "0 < degree f"
  shows "degree f - 1 \<in> keys f"
  using assms by transfer (fact degree_greater_zero_in_keys)

lemma in_keys_less_degree:
  "n \<in> keys f \<Longrightarrow> n < degree f"
  by transfer (fact in_keys_less_degree)


subsection \<open>Inductive structure\<close>

lift_definition update :: "nat \<Rightarrow> 'a \<Rightarrow> 'a :: zero nat_mapping \<Rightarrow> 'a nat_mapping"
  is "Poly_Mapping.update" .

lemma update_induct [case_names const update]:
  assumes const': "P 0"
  assumes update': "\<And>f a b. a \<notin> keys f \<Longrightarrow> b \<noteq> 0 \<Longrightarrow> P f \<Longrightarrow> P (update a b f)"
  shows "P f"
  using assms by transfer (erule update_induct, blast)

lemma lookup_update:
  "lookup (update k v f) k' = (if k = k' then v else lookup f k')"
  by transfer (fact lookup_update)

lemma keys_update:
  "keys (update k v f) = (if v = 0 then keys f - {k} else insert k (keys f))"
  by transfer (fact keys_update)


subsection \<open>Misc material\<close>

lemma inj_Abs_nat_mapping [simp]: "inj Abs_nat_mapping"
by(auto intro: injI simp add: Abs_nat_mapping_inject)

lemma inj_Rep_nat_mapping [simp]: "inj Rep_nat_mapping"
by(auto intro: injI simp add: Rep_nat_mapping_inject)

lemma vimage_Abs_nat_mapping: "Abs_nat_mapping -` A = Rep_nat_mapping ` A"
by(auto simp add: Rep_nat_mapping_inverse) (metis Abs_nat_mapping_inverse UNIV_I image_iff)

lift_definition nat_mapping_size :: "('a :: zero \<Rightarrow> nat) \<Rightarrow> 'a nat_mapping \<Rightarrow> nat"
  is "poly_mapping_size (\<lambda>n. n)" .

lemma nat_mapping_size_zero [simp]:
  "nat_mapping_size f 0 = f 0"
  by transfer simp

lemma nat_mapping_size_single [simp]:
  "nat_mapping_size f (single n v) = (if v = 0 then f 0 else f 0 + n + f v + 1)"
  by transfer simp

lemma less_nat_mapping_size:
  "n \<in> keys m \<Longrightarrow> n + f (lookup m n) < nat_mapping_size f m"
  by transfer (simp add: keys_less_poly_mapping_size [where f = "\<lambda>n. n"])

lemma nat_mapping_size_cong [fundef_cong]:
  "m = m' \<Longrightarrow> f 0 = f' 0 \<Longrightarrow> (\<And>v. v \<in> range m' \<Longrightarrow> f v = f' v)
  \<Longrightarrow> nat_mapping_size f m = nat_mapping_size f' m'"
  by transfer (rule poly_mapping_size_cong, simp_all)

lemma nat_mapping_size_estimation [termination_simp]:
  "k \<in> keys m \<Longrightarrow> y \<le> k + f (lookup m k) \<Longrightarrow> y < nat_mapping_size f m"
  by transfer (rule poly_mapping_size_estimation)

lemma nat_mapping_size_estimation2 [termination_simp]:
  "v \<in> range m \<Longrightarrow> y \<le> f v \<Longrightarrow> y < nat_mapping_size f m"
  by transfer (rule poly_mapping_size_estimation2)

lemma lookup_le_nat_mapping_size [termination_simp]:
  "f (lookup m k) \<le> nat_mapping_size f m"
  by transfer (rule lookup_le_poly_mapping_size)

lift_definition mapp :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero) \<Rightarrow> 'a nat_mapping \<Rightarrow> 'b nat_mapping"
  is Poly_Mapping.mapp .

lemma lookup_mapp: "lookup (mapp f p) k = (f k (lookup p k) when (k \<in> keys p))"
by transfer(simp add: lookup_mapp)

(* We assume inc_power hidden from PP-users, thus it is not defined in Poly_Mapping.thy
  FIXME: is the following appropriate..? *)
definition inc_power :: "nat \<Rightarrow> nat nat_mapping \<Rightarrow> nat nat_mapping"
where
  "inc_power n p = (if n \<in> keys p then (update n (Suc (lookup p n)) p) else 0)"


definition sum_nat_mapping :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: comm_monoid_add) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b"
where "sum_nat_mapping f p = setsum (\<lambda>k. f k (Nat_Mapping.lookup p k)) (Nat_Mapping.keys p)"

lemma sum_nat_mapping_zero [simp]:
  "sum_nat_mapping f 0 = 0"
by(simp add: sum_nat_mapping_def)

lemma sum_nat_mapping_one [simp]:
  "sum_nat_mapping f 1 = f 0 1"
by(simp add: sum_nat_mapping_def)

lemma sum_nat_mapping_single [simp]:
  "sum_nat_mapping f (Nat_Mapping.single k v) = (f k v when v \<noteq> 0)"
by(simp add: sum_nat_mapping_def)

lemma sum_nat_mapping_cong [fundef_cong]:
  "\<lbrakk> m = m'; \<And>k. k \<in> Nat_Mapping.keys m' \<Longrightarrow> f k (Nat_Mapping.lookup m' k) = f' k (Nat_Mapping.lookup m' k) \<rbrakk>
  \<Longrightarrow> sum_nat_mapping f m = sum_nat_mapping f' m'"
by(simp add: sum_nat_mapping_def)

lemma sum_nat_mapping_plus_distrib:
  assumes "\<And>k. f k 0 = 0"
  and "\<And>k a b. k \<in> Nat_Mapping.keys p \<union> Nat_Mapping.keys q \<Longrightarrow> f k (Nat_Mapping.lookup p k + Nat_Mapping.lookup q k) = f k (Nat_Mapping.lookup p k) + f k (Nat_Mapping.lookup q k)"
  shows "sum_nat_mapping f (p + q) = sum_nat_mapping f p + sum_nat_mapping f q"
unfolding sum_nat_mapping_def using assms
apply(transfer)
apply(subst setsum_keys_plus_distrib)
apply simp_all
done

lemma keys_mapp_subset: "keys (mapp f p) \<subseteq> keys p"
by transfer(rule keys_mapp_subset)

lemma sum_nat_mapping_mapp:
  assumes hom_0: "\<And>k. f k 0 = 0"
  shows "sum_nat_mapping f (mapp g p) = sum_nat_mapping (\<lambda>n x. f n (g n x)) p"
apply(simp add: sum_nat_mapping_def lookup_mapp)
apply(rule setsum.mono_neutral_cong_left)
using keys_mapp_subset by(auto simp add: lookup_mapp hom_0 when_def)

lemma sum_nat_mapping_K_zero [simp]:
  "sum_nat_mapping (\<lambda>_ _. 0) p = 0"
by(simp add: sum_nat_mapping_def)

lemma sum_nat_mapping_right_distrib:
  fixes c :: "_ :: semiring_0" shows
  "c * sum_nat_mapping f p = sum_nat_mapping (\<lambda>n x. c * f n x) p"
by(simp add: sum_nat_mapping_def setsum_right_distrib)

subsection \<open>Dense implementation\<close>

subsubsection \<open>Morphisms with lists\<close>

typedef (overloaded) 'a :: zero nat_mapping_dense = "{xs :: 'a list. no_trailing_zeros xs}"
  by auto

setup_lifting type_definition_nat_mapping_dense

lemma no_trailing_zeros_Rep_nat_mapping_dense [simp]:
  "no_trailing_zeros (Rep_nat_mapping_dense m)"
  using Rep_nat_mapping_dense [of m] by simp

lemma strip_while_Rep_nat_mapping_dense:
  "strip_while (op = 0) (Rep_nat_mapping_dense m) = Rep_nat_mapping_dense m"
  by (simp add: strip_while_idem_iff)

lemma strip_while_Rep_nat_mapping_dense':
  "strip_while (\<lambda>x. x = 0) (Rep_nat_mapping_dense m) = Rep_nat_mapping_dense m"
  by (subst eq_commute) (fact strip_while_Rep_nat_mapping_dense)


subsubsection \<open>Morphisms with polynomial mappings\<close>

lift_definition nth :: "'a list \<Rightarrow> \<nat> \<Rightarrow>\<^sub>0 'a::zero"
  is "Poly_Mapping.nth" .

definition "values" :: "(\<nat> \<Rightarrow>\<^sub>0 'a::zero) \<Rightarrow> 'a list"
where
  "values f = map (lookup f) [0..<degree f]"

lemma no_trailing_zeros_values:
  "no_trailing_zeros (values f)"
  by (auto simp add: values_def no_trailing_map dest: degree_greater_zero_in_keys)

lemma nth_values:
  "nth (values f) = f"
  unfolding values_def by transfer (fact nth_idem)

lemma values_nth:
  "no_trailing_zeros xs \<Longrightarrow> values (nth xs) = xs"
  unfolding values_def by transfer (simp add: degree_nth map_nth_default)

lemma values_zero [simp]:
  "values 0 = []"
  unfolding values_def by simp

lemma values_one [simp]:
  "values 1 = [1]"
  unfolding values_def by simp

lemma values_single [simp]:
  "values (single k v) = (if v = 0 then [] else replicate k 0 @ [v])"
  unfolding values_def by (auto intro!: replicate_eqI)


subsubsection \<open>Properties\<close>

lift_definition single_dense :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'a nat_mapping_dense"
  is "\<lambda>n c. if c = 0 then [] else replicate n 0 @ [c]"
  by (simp split: split_if_asm)

definition apply0 :: "('a :: zero \<Rightarrow> 'b :: zero) \<Rightarrow> 'a \<Rightarrow> 'b" (infixl "0$" 100)
where "apply0 f x = (if x = 0 then 0 else f x)"

lemma apply0_0 [simp]: "f 0$ 0 = 0"
by(simp add: apply0_def)

lemma apply0_neq0 [simp]: "x \<noteq> 0 \<Longrightarrow> f 0$ x = f x"
by(simp add: apply0_def)

context
  fixes f :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero"
begin

fun strip_map :: "nat \<Rightarrow> 'a list \<Rightarrow> 'b list"
where
  "strip_map n [] = []"
| "strip_map n (x#xs) =
  (let ys = strip_map (Suc n) xs; y = f n 0$ x in if ys = [] \<and> y = 0 then [] else y # ys)"

lemma no_trailing_zero_strip_map [simp]:
  "no_trailing_zeros (strip_map n xs)"
by(induction n xs rule: strip_map.induct)(simp_all add: Let_def)

lemma nth_strip_map [simp]:
  "i < length (strip_map n xs)
  \<Longrightarrow> strip_map n xs ! i = f (n + i) 0$ (nth_default 0 xs i)"
by(induction n xs arbitrary: i rule: strip_map.induct)(auto simp add: Let_def nth_default_Cons split: split_if_asm nat.splits)

lemma nth_default_strip_map [simp]:
  "nth_default 0 (strip_map n xs) i = f (n + i) 0$ (nth_default 0 xs i)"
by(induction n xs arbitrary: i rule: strip_map.induct)(auto simp add: Let_def nth_default_Cons split: split_if_asm nat.splits)

end

definition apply0_2 :: "('a :: zero \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero) \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c"
where "apply0_2 f x y = (if x = 0 \<and> y = 0 then 0 else f x y)"

lemma apply0_2_0_0 [simp]: "apply0_2 f 0 0 = 0"
by(simp add: apply0_2_def)

lemma apply0_2_neq0 [simp]: "x \<noteq> 0 \<or> y \<noteq> 0 \<Longrightarrow> apply0_2 f x y = f x y"
by(auto simp add: apply0_2_def)

lemma apply0_2_0: "apply0_2 f 0 x = f 0 0$ x"
by(simp add: apply0_2_def)

lemma apply0_2_0': "apply0_2 f x 0 = (\<lambda>x. f x 0) 0$ x"
by(simp add: apply0_2_def)

context
  fixes f :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero"
begin

function zip_with_dense_raw :: "nat \<Rightarrow> 'a list \<Rightarrow> 'b list \<Rightarrow> 'c list"
where
  "zip_with_dense_raw n [] ys = strip_map (\<lambda>n y. f n 0 y) n ys"
| "zip_with_dense_raw n xs [] = strip_map (\<lambda>n x. f n x 0) n xs"
| "zip_with_dense_raw n (x#xs) (y#ys) =
  (let zs = zip_with_dense_raw (Suc n) xs ys; z = apply0_2 (f n) x y in if zs = [] \<and> z = 0 then [] else z # zs)"
by pat_completeness auto
termination by lexicographic_order

context
  fixes g :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero"
begin

fun mapp_dense_raw :: "nat \<Rightarrow> 'a list \<Rightarrow> 'b list"
where
  "mapp_dense_raw n [] = []"
| "mapp_dense_raw n (x#xs) = 
  (let zs = mapp_dense_raw (Suc n) xs; z = apply0 (g n) x in if zs = [] \<and> z = 0 then [] else z # zs)"

end

declare [[simproc del: Groups.reorient_zero]]

lemma no_trailing_zeros_zip_with_dense_raw [simp]:
  "no_trailing_zeros (zip_with_dense_raw n xs ys)"
by(induct xs ys rule: zip_with_dense_raw.induct)(auto simp add: Let_def)

declare [[simproc add: Groups.reorient_zero]]

lemma nth_zip_with_dense_raw [simp]:
  "i < length (zip_with_dense_raw n xs ys)
  \<Longrightarrow> zip_with_dense_raw n xs ys ! i =
     apply0_2 (f (n + i)) (nth_default 0 xs i) (nth_default 0 ys i)"
apply(induction n xs ys arbitrary: i rule: zip_with_dense_raw.induct)
apply(auto simp add: Let_def nth_default_Cons split: split_if_asm nat.split)
apply(simp_all add: apply0_2_def)
done

lemma nth_default_zip_with_dense_raw [simp]:
  "nth_default 0 (zip_with_dense_raw n xs ys) i
  = apply0_2 (f (n + i)) (nth_default 0 xs i) (nth_default 0 ys i)"
apply(induction n xs ys arbitrary: i rule: zip_with_dense_raw.induct)
apply(auto simp add: Let_def nth_default_Cons split: split_if_asm nat.split)
apply(simp_all add: apply0_2_def)
done

end

lift_definition zip_with_dense
  :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero)
  \<Rightarrow> 'a nat_mapping_dense \<Rightarrow> 'b nat_mapping_dense \<Rightarrow> 'c nat_mapping_dense"
  is "\<lambda>f. zip_with_dense_raw f 0"
by simp

lift_definition mapp_dense
  :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero)
  \<Rightarrow> 'a nat_mapping_dense \<Rightarrow> 'b nat_mapping_dense"
  is "\<lambda>f. mapp_dense_raw f 0"
sorry

instantiation nat_mapping_dense :: ("{equal, zero}") equal begin
lift_definition equal_nat_mapping_dense :: "'a nat_mapping_dense \<Rightarrow> 'a nat_mapping_dense \<Rightarrow> bool" 
  is "op =" .
instance by(intro_classes)(transfer, simp)
end

instantiation nat_mapping_dense :: ("{linorder, zero}") linorder begin
lift_definition less_nat_mapping_dense :: "'a nat_mapping_dense \<Rightarrow> 'a nat_mapping_dense \<Rightarrow> bool" 
  is "op <" .
lift_definition less_eq_nat_mapping_dense :: "'a nat_mapping_dense \<Rightarrow> 'a nat_mapping_dense \<Rightarrow> bool" 
  is "op \<le>" .
instance
  by(intro_classes)(transfer,
    (rule less_le_not_le linear order_refl|erule (1) order_trans order_antisym))+
end

lift_definition lookup_dense :: "'a :: zero nat_mapping_dense \<Rightarrow> nat \<Rightarrow> 'a"
is "nth_default 0" .

lift_definition range_dense :: "'a :: zero nat_mapping_dense \<Rightarrow> 'a set"
is "\<lambda>xs. set xs - {0}" .

context
  fixes f :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b"
begin

fun foldr_dense_raw :: "nat \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b"
where
  "foldr_dense_raw n []     b = b"
| "foldr_dense_raw n (x#xs) b = (if x = 0 then id else f n x) (foldr_dense_raw (Suc n) xs b)"

end

lemma foldr_dense_raw_conv_foldr:
  "foldr_dense_raw f n xs z = List.foldr (\<lambda>(k, v) r. if v \<noteq> 0 then f k v r else r)
    (enumerate n xs) z"
proof(induction n xs z rule: foldr_dense_raw.induct)
  case 2 thus ?case unfolding enumerate_eq_zip by (subst upt_rec) simp
qed simp

lift_definition foldr_dense :: "(nat \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a :: zero nat_mapping_dense \<Rightarrow> 'b \<Rightarrow> 'b"
is "\<lambda>f. foldr_dense_raw f 0" .

context
  fixes dflt :: 'a
begin

fun extend_with_update :: "nat \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "extend_with_update 0       a []       = [a]"
| "extend_with_update (Suc n) a []       = dflt # extend_with_update n a []"
| "extend_with_update 0       a (x # xs) = a # xs"
| "extend_with_update (Suc n) a (x # xs) = x # extend_with_update n a xs"

end

definition update_dense_raw :: "nat \<Rightarrow> 'a \<Rightarrow> 'a :: zero list \<Rightarrow> 'a list"
where "update_dense_raw n a xs = strip_while (op = 0) (extend_with_update 0 n a xs)"

lift_definition update_dense :: "nat \<Rightarrow> 'a \<Rightarrow> 'a :: zero nat_mapping_dense \<Rightarrow> 'a nat_mapping_dense"
is "update_dense_raw" by(simp add: update_dense_raw_def)


subsection \<open>Sparse implementation\<close>

subsubsection \<open>Morphisms with association lists\<close>

typedef (overloaded) 'a :: zero nat_mapping_sparse = 
  "{xs :: (nat \<times> 'a) list. sorted (map fst xs) \<and> distinct (map fst xs) \<and> 0 \<notin> snd ` set xs}"
  by (auto intro: exI [of _ "[]"])

setup_lifting type_definition_nat_mapping_sparse


subsubsection \<open>Properties\<close>

lift_definition single_sparse :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'a nat_mapping_sparse"
is "\<lambda>n c. if c = 0 then [] else [(n, c)]" by auto

context fixes f :: "'a :: linorder \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero \<Rightarrow> 'd :: zero" begin

definition zip_with_sparse_raw :: "('a \<times> 'b) list \<Rightarrow> ('a \<times> 'c) list \<Rightarrow> ('a \<times> 'd) list"
where 
  "zip_with_sparse_raw xs ys =
   oa_zip_with' comparator_of_le (\<lambda>k v. let z = f k v 0 in if z = 0 then None else Some z) 
    (\<lambda>k v. let z = f k 0 v in if z = 0 then None else Some z) 
    (\<lambda>k v v'. let z = f k v v' in if z = 0 then None else Some z) xs ys"

lemma set_zip_with_sparse_raw:
  assumes "sorted (map fst xs)" "sorted (map fst ys)" "distinct (map fst xs)" "distinct (map fst ys)"
  shows 
  "set (zip_with_sparse_raw xs ys) = 
   (\<Union>x\<in>fst ` set xs \<union> fst ` set ys. 
    Pair x ` (case (c.lookup comparator_of_le xs x, c.lookup comparator_of_le ys x) of
       (Some v, None) \<Rightarrow> (if f x v 0 = 0 then {} else {f x v 0})
     | (None, Some v') \<Rightarrow> (if f x 0 v' = 0 then {} else {f x 0 v'})
     | (Some v, Some v') \<Rightarrow> (if f x v v' = 0 then {} else {f x v v'})))"
unfolding zip_with_sparse_raw_def
by(subst set_oa_zip_with')(simp_all add: dsorted_by_conv_sorted_distinct assms dsorted_by_vimage Let_def if_distrib[where f=set_option] cong: option.case_cong if_cong)

lemma 
  assumes "sorted (map fst xs)" "sorted (map fst ys)" "distinct (map fst xs)" "distinct (map fst ys)"
  shows sorted_zip_with_sparse_raw: "sorted (map fst (zip_with_sparse_raw xs ys))" (is ?sorted)
  and distinct_zip_with_sparse_raw: "distinct (map fst (zip_with_sparse_raw xs ys))" (is ?distinct)
proof -
  from assms 
  have "c.dsorted_by comparator_of_le (map fst xs)" "c.dsorted_by comparator_of_le (map fst ys)"
    by(simp_all add: sorted_distinct_dsorted_by)
  with _ have "c.dsorted_by comparator_of_le (map fst (zip_with_sparse_raw xs ys))"
    unfolding zip_with_sparse_raw_def by(rule dsorted_by_oa_zip_with') simp
  thus "?sorted" "?distinct"
    by(simp_all add: dsorted_by_conv_sorted_distinct)
qed

lemma lookup_zip_with_sparse_raw:
  assumes "sorted (map fst xs)" "sorted (map fst ys)" "distinct (map fst xs)" "distinct (map fst ys)"
  shows
  "c.lookup comparator_of_le (zip_with_sparse_raw xs ys) k =
  (case (c.lookup comparator_of_le xs k, c.lookup comparator_of_le ys k) of
     (None, None) \<Rightarrow> None
   | (Some v, None) \<Rightarrow> (if f k v 0 = 0 then None else Some (f k v 0))
   | (None, Some v') \<Rightarrow> (if f k 0 v' = 0 then None else Some (f k 0 v'))
   | (Some v, Some v') \<Rightarrow> if f k v v' = 0 then None else Some (f k v v'))"
by(simp add: zip_with_sparse_raw_def lookup_oa_zip_with' Let_def dsorted_by_conv_sorted_distinct dsorted_by_vimage assms cong: if_cong option.case_cong)

end

lift_definition zip_with_sparse
  :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero) \<Rightarrow> 
    'a nat_mapping_sparse \<Rightarrow> 'b nat_mapping_sparse \<Rightarrow> 'c nat_mapping_sparse"
is "zip_with_sparse_raw"
by(force split: option.splits split_if_asm simp add: set_zip_with_sparse_raw comparator_eq_on.lookup_eq_None_iff[where D=UNIV] comparator_eq_on.lookup_eq_Some_iff[where D=UNIV] dsorted_by_vimage Let_def dsorted_by_conv_sorted_distinct dsorted_by_vimage sorted_zip_with_sparse_raw distinct_zip_with_sparse_raw intro: rev_image_eqI)

context
  fixes g :: "'a :: linorder \<Rightarrow> 'b :: zero \<Rightarrow> 'c :: zero" (* g preserves order *)
begin

definition mapp_sparse_raw :: "('a \<times> 'b) list \<Rightarrow> ('a \<times> 'c) list"
where 
  "mapp_sparse_raw xs = map (\<lambda>(k, v). (k, g k v)) xs"

end

lift_definition mapp_sparse 
  :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b :: zero)
  \<Rightarrow> 'a nat_mapping_sparse \<Rightarrow> 'b nat_mapping_sparse"
  is "mapp_sparse_raw"
sorry

instantiation nat_mapping_sparse :: ("{zero, equal}") equal begin
lift_definition equal_nat_mapping_sparse :: "'a nat_mapping_sparse \<Rightarrow> 'a nat_mapping_sparse \<Rightarrow> bool" 
  is "op = :: (nat \<times> 'a) list \<Rightarrow> _" .
instance by(intro_classes)(transfer, simp)
end

instantiation nat_mapping_sparse :: ("{zero, linorder}") linorder begin
lift_definition less_nat_mapping_sparse :: "'a nat_mapping_sparse \<Rightarrow> 'a nat_mapping_sparse \<Rightarrow> bool"
  is "op < :: (nat \<times> 'a) list \<Rightarrow> _" .
lift_definition less_eq_nat_mapping_sparse 
  :: "'a nat_mapping_sparse \<Rightarrow> 'a nat_mapping_sparse \<Rightarrow> bool"
  is "op \<le> :: (nat \<times> 'a) list \<Rightarrow> _" .
instance
  by(intro_classes)
    (transfer, (rule less_le_not_le linear order_refl|erule (1) order_trans order_antisym))+
end

lift_definition lookup_sparse :: "'a :: zero nat_mapping_sparse \<Rightarrow> nat \<Rightarrow> 'a"
is "\<lambda>m x. case map_of m x of None \<Rightarrow> 0 | Some v \<Rightarrow> v" .

lift_definition range_sparse :: "'a :: zero nat_mapping_sparse \<Rightarrow> 'a set"
is "\<lambda>xs. set (map snd xs)" .

lift_definition foldr_sparse :: "(nat \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a :: zero nat_mapping_sparse \<Rightarrow> 'b \<Rightarrow> 'b"
is "\<lambda>f. List.foldr (\<lambda>(k, v). f k v)" .

lift_definition update_sparse :: "nat \<Rightarrow> 'a :: zero \<Rightarrow> 'a nat_mapping_sparse \<Rightarrow> 'a nat_mapping_sparse"
is "\<lambda>n c (xs :: (nat * 'a) list). if c = 0 then c.remove_by (compare_vimage fst comparator_of_le) (n, c) xs else c.update_by comparator_of_le n c xs"
apply(fold conj_assoc dsorted_by_conv_sorted_distinct dsorted_by_vimage)
apply(auto simp add: dsorted_by_update_by[where D=UNIV] comparator_on.dsorted_by_remove_by[where D=UNIV] dest: c.in_set_remove_byD c.in_set_update_by_weakD)
done


subsection \<open>Connection with @{typ "_ nat_mapping"}\<close>

lemma mapp_cong [fundef_cong]:
  "\<lbrakk> m = m'; \<And>k. k \<in> keys m' \<Longrightarrow> f k (lookup m' k) = f' k (lookup m' k)\<rbrakk>
  \<Longrightarrow> mapp f m = mapp f' m'"    
  by transfer (rule Poly_Mapping.mapp_cong)

lift_definition Dense :: "'a :: zero nat_mapping_dense \<Rightarrow> 'a nat_mapping"
  is nth .

lift_definition rep_dense :: "'a :: zero nat_mapping \<Rightarrow> 'a nat_mapping_dense"
  is "values" by (fact no_trailing_zeros_values)

lemma Dense_rep_dense: "Dense (rep_dense p) = p"
  by transfer (fact nth_values)

lemma rep_dense_Dense: "rep_dense (Dense xs) = xs"
  by transfer (fact values_nth)

lemma iso_nat_mapping_dense: "type_definition rep_dense Dense UNIV"
  by unfold_locales (simp_all add: Dense_rep_dense rep_dense_Dense)

context begin
interpretation iso: type_definition rep_dense Dense UNIV
by(rule iso_nat_mapping_dense)

lemma Dense_inject: "Dense xs = Dense ys \<longleftrightarrow> xs = ys"
by(rule iso.Abs_inject) simp_all

lemma rep_dense_inject: "rep_dense p = rep_dense q \<longleftrightarrow> p = q"
by(rule iso.Rep_inject)

end

lift_definition Sparse :: "'a nat_mapping_sparse \<Rightarrow> \<nat> \<Rightarrow>\<^sub>0 'a :: zero"
  is "of_oalist comparator_of_le" .

lift_definition rep_sparse :: "(\<nat> \<Rightarrow>\<^sub>0 'a :: zero) \<Rightarrow> 'a nat_mapping_sparse"
  is "to_oalist comparator_of_le"
  by auto

lemma Sparse_rep_sparse: "Sparse (rep_sparse p) = p"
  by transfer simp
  
lemma rep_sparse_Sparse: "rep_sparse (Sparse p) = p"
by transfer(simp add: to_oalist_comparator_of_le[symmetric] dsorted_by_conv_sorted_distinct dsorted_by_vimage)

lemma iso_nat_mapping_sparse: "type_definition rep_sparse Sparse UNIV"
by unfold_locales(simp_all add: Sparse_rep_sparse rep_sparse_Sparse)

context begin
interpretation iso: type_definition rep_sparse Sparse UNIV
by(rule iso_nat_mapping_sparse)

lemma Sparse_inject: "Sparse p = Sparse q \<longleftrightarrow> p = q"
by(rule iso.Abs_inject) simp_all

lemma rep_sparse_inject: "rep_sparse p = rep_sparse q \<longleftrightarrow> p = q"
by(fact iso.Rep_inject)

end


text \<open>A type to refer to representation of @{typ "_ nat_mapping"} in the code\<close>

datatype impl = IMPL

lemma impl_eq [no_atp]: "x = IMPL" by(cases x)(simp)

text \<open>
  Simplification procedure for @{thm [source] impl_eq}.  Cannot use
  this rule directly --- it loops!
\<close>

simproc_setup impl_eq ("x::impl") = \<open>fn _ => fn _ => fn ct =>
  case Thm.term_of ct of Const (@{const_name IMPL}, _) => NONE | _ => SOME (mk_meta_eq @{thm impl_eq})
\<close>

lemma equal_impl_code [code]:
  "HOL.equal (x :: impl) (y :: impl) = True"
by(simp add: equal_eq)

definition DENSE :: impl where [code del, simp]: "DENSE = IMPL"
definition SPARSE :: impl where [code del, simp]: "SPARSE = IMPL"

code_datatype DENSE SPARSE

definition implT :: "'a :: zero nat_mapping \<Rightarrow> impl"
where [code del, simp]: "implT _ = IMPL"

hide_const (open) implT

adhoc_overloading implT Nat_Mapping.implT

lemma implT_code [code]:
  "implT (Dense d) = DENSE"
  "implT (Sparse s) = SPARSE"
by simp_all


lemma less_length_strip_whileI: "\<lbrakk> x < length xs; \<not> P (xs ! x) \<rbrakk> \<Longrightarrow> x < length (strip_while P xs)"
by(induct xs rule: rev_induct)(simp_all add: nth_append strip_while_def split: split_if_asm)

lemma nth_Nil [simp]: "nth [] = 0"
  by transfer simp

lemma nth_singleton [simp]: "nth [v] = single 0 v"
  by transfer simp

definition of_list :: "impl \<Rightarrow> 'a :: zero list \<Rightarrow> 'a nat_mapping"
where
  [simp, code del]: "of_list _ = nth"

lift_definition of_list_dense :: "'a :: zero list \<Rightarrow> 'a nat_mapping_dense"
is "\<lambda>xs. strip_while (op = 0) xs" by simp

lemma of_list'_DENSE [code]:
  "of_list DENSE xs = Dense (of_list_dense xs)"
  apply transfer apply simp apply transfer apply simp done

definition of_list_sparse_impl :: "'a :: zero list \<Rightarrow> (nat \<times> 'a) list"
where "of_list_sparse_impl xs = filter (\<lambda>(k, v). v \<noteq> 0) (zip [0..<length xs] xs)"

lift_definition of_list_sparse :: "'a :: zero list \<Rightarrow> 'a nat_mapping_sparse"
is "of_list_sparse_impl"
by(auto simp add: of_list_sparse_impl_def sorted_filter distinct_map_filter split_def)

lemma of_list'_SPARSE [code]:
  "of_list SPARSE xs = Sparse (of_list_sparse xs)"
unfolding of_list_def
apply transfer
apply transfer
apply(simp add: of_list_sparse_impl_def fun_eq_iff c.lookup_default_alt_def[abs_def] lookup_conv_map_of'[OF comparator_eq_on_comparator_of_le] dsorted_by_conv_sorted_distinct dsorted_by_vimage distinct_map_filter sorted_filter)
apply(auto split: option.split simp add: map_of_eq_None_iff
  image_Collect distinct_map_filter enumerate_eq_zip [of 0, symmetric, simplified]
  in_set_enumerate_eq nth_default_beyond nth_default_nth nth_default_eq_dflt_iff)
done

code_datatype Dense Sparse

instantiation nat_mapping :: ("{zero, equal}") equal begin
definition equal_nat_mapping :: "'a nat_mapping \<Rightarrow> 'a nat_mapping \<Rightarrow> bool"
where "equal_nat_mapping = op ="
instance by(intro_classes)(simp add: equal_nat_mapping_def)
end

lemma equal_nat_mapping_code [code]:
  "\<And>x y :: _ nat_mapping. HOL.equal x y \<longleftrightarrow>
   Code.abort (STR ''equality comparison between different representations unsupported'')
    (\<lambda>_. HOL.equal x y)"
  "\<And>x y. HOL.equal (Dense x) (Dense y) \<longleftrightarrow> x = y"
  "\<And>x y. HOL.equal (Sparse x) (Sparse y) \<longleftrightarrow> x = y"
defer unfolding HOL.equal_eq by(rule Dense_inject Sparse_inject)+ simp

lemma less_nat_mapping_code [code]:
  "\<And>x y :: _ nat_mapping. x < y \<longleftrightarrow>
   Code.abort (STR ''strict order comparison between different representations unsupported'')
    (\<lambda>_. x < y)"
  "\<And>x y. Dense x < Dense y \<longleftrightarrow> x < y"
  "\<And>x y. Sparse x < Sparse y \<longleftrightarrow> x < y"
sorry

lemma less_eq_nat_mapping_code [code]:
  "\<And>x y :: _ nat_mapping. x \<le> y \<longleftrightarrow>
   Code.abort (STR ''order comparison between different representations unsupported'')
    (\<lambda>_. x \<le> y)"
  "\<And>x y. Dense x \<le> Dense y \<longleftrightarrow> x \<le> y"
  "\<And>x y. Sparse x \<le> Sparse y \<longleftrightarrow> x \<le> y"
sorry

lemma lookup_nat_mapping_code [code]:
  "\<And>x. lookup (Dense x) = lookup_dense x"
  "\<And>x. lookup (Sparse x) = lookup_sparse x"
sorry

lemma length_zip_with_dense_raw_estimation:
  "\<lbrakk> i < max (length xs) (length ys); f (n + i) (nth_default 0 xs i) (nth_default 0 ys i) \<noteq> 0 \<rbrakk>
  \<Longrightarrow> i < length (zip_with_dense_raw f n xs ys)"
sorry

lemma apply0_2_eq_0_iff:
  "apply0_2 f x y = 0 \<longleftrightarrow> (x \<noteq> 0 \<or> y \<noteq> 0 \<longrightarrow> f x y = 0)"
by(auto simp add: apply0_2_def)

lemma plus_Dense_Dense:
  "Dense xs + Dense ys = Dense (zip_with_dense (\<lambda>_. plus) xs ys)"
proof transfer
  fix xs ys :: "'a list"
  show "nth xs + nth ys = nth (zip_with_dense_raw (\<lambda>_. plus) 0 xs ys)"
  proof transfer
      fix xs ys :: "'a list"
    { fix xs ys :: "'a list"
      have "finite {n. nth_default 0 xs n \<noteq> 0 \<or> nth_default 0 ys n \<noteq> 0}"
        unfolding finite_Collect_disjI by simp_all
    } note * = this
    show "Poly_Mapping.nth xs + Poly_Mapping.nth ys =
      Poly_Mapping.nth (zip_with_dense_raw (\<lambda>_. plus) 0 xs ys)"
        by transfer (simp add: fun_eq_iff apply0_2_def)
  qed
qed

lemma plus_Sparse_Sparse:
  "Sparse xs + Sparse ys = Sparse (zip_with_sparse (\<lambda>_. plus) xs ys)"
proof transfer
  fix xs ys :: "(nat \<times> 'a) list"
  assume xs: "sorted (map fst xs) \<and> distinct (map fst xs) \<and> 0 \<notin> snd ` set xs"
  assume ys: "sorted (map fst ys) \<and> distinct (map fst ys) \<and> 0 \<notin> snd ` set ys"
  have *: "\<And>k. (if comparator_on_base.lookup_default comparator_of_le 0 xs k = 0
   \<longrightarrow> comparator_on_base.lookup_default comparator_of_le 0 ys k \<noteq> 0
     then comparator_on_base.lookup_default comparator_of_le 0 xs k +
       comparator_on_base.lookup_default comparator_of_le 0 ys k
     else 0) =
     (comparator_on_base.lookup_default comparator_of_le 0
       (zip_with_sparse_raw (\<lambda>_. plus) xs ys)) k"
    apply (subst (5) c.lookup_default_alt_def)
    apply simp
    apply (simp add: lookup_zip_with_sparse_raw xs ys)
    apply (auto split del: split_if split: option.split)
    apply (simp_all add: c.lookup_default_alt_def split: split_if_asm)
    done
  show "of_oalist comparator_of_le xs + 
   of_oalist comparator_of_le ys = of_oalist comparator_of_le (zip_with_sparse_raw (\<lambda>_. plus) xs ys)"
    apply (rule poly_mapping_eqI)
    apply (simp add: Poly_Mapping.lookup_add)
    apply (simp add: lookup_of_oalist_comparator_of_le sorted_zip_with_sparse_raw distinct_zip_with_sparse_raw xs ys)
    apply (simp add: * [symmetric])
    done
qed

lemma mapp_Dense: 
  "Nat_Mapping.mapp f (Dense xs) = Dense (mapp_dense f xs)"
sorry

lemma mapp_Sparse:
  "Nat_Mapping.mapp f (Sparse xs) = Sparse (mapp_sparse f xs)"
sorry

lemmas plus_code [code] = 
  plus_Dense_Dense
  plus_Sparse_Sparse

lemmas cmult_code [code] =
  mapp_Dense
  mapp_Sparse


definition single' :: "impl \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a :: zero nat_mapping"
where [code del, simp]: "single' _ = single"

lemma single'_DENSE:
  "single' DENSE k v = Dense (single_dense k v)"
apply transfer
apply auto
apply transfer
apply transfer
apply (auto simp add: fun_eq_iff nth_default_def nth_append)
done


lemma single'_SPARSE:
  "single' SPARSE k v = Sparse (single_sparse k v)"
unfolding single'_def
by transfer(simp add: of_oalist_single)

lemmas single'_code [code] =
  single'_DENSE
  single'_SPARSE

lemma range_Dense:
  "range (Dense xs) = range_dense xs"
  apply transfer
  apply transfer
  apply simp
  done

lemma range_Sparse: "range (Sparse xs) = range_sparse xs"
apply transfer
apply transfer
apply(auto simp add: range_lookup_default ran_def comparator_eq_on.lookup_eq_Some_iff[OF comparator_eq_on_comparator_of_le, where D=UNIV] dsorted_by_conv_sorted_distinct dsorted_by_vimage intro: rev_image_eqI)
done

lemmas range_code [code] =
  range_Dense
  range_Sparse

lemma update_Dense: "update n a (Dense xs) = Dense (update_dense n a xs)"
proof (transfer, transfer, transfer)
  fix n a and xs :: "'a list"
  have "nth_default 0 (update_dense_raw n a xs) = (nth_default 0 xs)(n := a)"
    unfolding update_dense_raw_def nth_default_strip_while_dflt
    apply (induct n _ xs rule: extend_with_update.induct)
    apply (auto simp add: fun_eq_iff gr0_conv_Suc)
    apply (case_tac x) apply simp_all
    apply (case_tac xa) apply simp_all
    done
  then show "(nth_default 0 xs)(n := a) = nth_default 0 (update_dense_raw n a xs)"
    by simp
qed

lemma update_Sparse: "update n a (Sparse xs) = Sparse (update_sparse n a xs)"
  apply transfer
  apply (subst of_oalist_comparator_of_le_conv_the_value) apply simp_all
  apply (subst of_oalist_comparator_of_le_conv_the_value) apply simp_all
  prefer 3
  apply auto
  sorry

lemmas update_code [code] =
  update_Dense
  update_Sparse


hide_type (open) impl
hide_const (open) single lookup keys range update nth "values" degree IMPL of_list

end