(* Author: Andreas Lochbihler, ETH Zurich
   Author: Florian Haftmann, TU Muenchen 
*)

section \<open>Particularties of nat mapping for recursive representation only\<close>

theory Nat_Mapping_Recursive
imports Nat_Mapping
begin

subsection \<open>A mere bootstrap device which does not occur in the lemmas-to-work-with\<close>

definition zip :: "('a::zero \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a)"
where
  "zip f g h = Nat_Mapping.nth (map (\<lambda>n. f (Nat_Mapping.lookup g n) (Nat_Mapping.lookup h n)) [0..<max (Nat_Mapping.degree g) (Nat_Mapping.degree h)])"

lemma zip_plus_plus:
  "zip plus = plus"
  unfolding zip_def [abs_def] fun_eq_iff
  by transfer (simp add: Poly_Mapping.lookup_add [symmetric] nth_idem_bound degree_add)

lemma zip_cong [fundef_cong]:
  "\<lbrakk> p = p'; q = q'; \<And>k. k \<in> Nat_Mapping.keys p' \<union> Nat_Mapping.keys q' \<Longrightarrow>
    f (Nat_Mapping.lookup p' k) (Nat_Mapping.lookup q' k) = f' (Nat_Mapping.lookup p' k) (Nat_Mapping.lookup q' k) \<rbrakk>
  \<Longrightarrow> zip f p q = zip f' p' q'"
  unfolding zip_def
  apply transfer
  unfolding nth_eq_iff
  unfolding strip_while_map comp_def
  apply (rule map_cong)
  apply auto
  sorry


subsection \<open>Items folding\<close>

lift_definition items :: "(\<nat> \<Rightarrow>\<^sub>0 'a::zero) \<Rightarrow> (nat \<times> 'a) list"
  is Poly_Mapping.items .

lemma items_unfold:
  "items f = [(k, v)\<leftarrow>map (\<lambda>k. (k, Nat_Mapping.lookup f k)) [0..<Nat_Mapping.degree f]. v \<noteq> 0]"
  apply transfer unfolding Poly_Mapping.items_def sorted_list_of_set_keys
  unfolding Poly_Mapping.degree_def
  apply transfer
  apply (simp add: filter_map comp_def)
  done

lemma items_unfold':
  "items f = [(k, v)\<leftarrow>enumerate 0 (Nat_Mapping.values f). v \<noteq> 0]"
  unfolding items_unfold values_def enumerate_map_upt ..

lemma items_zero [simp]:
  "items 0 = []"
  unfolding items_unfold by simp

lemma items_one [simp]:
  "items 1 = [(0, 1)]"
  unfolding items_unfold by simp

lemma items_single [simp]:
  "items (Nat_Mapping.single k v) = (if v = 0 then [] else [(k, v)])"
  unfolding items_unfold by simp

lemma in_set_items_iff [simp]:
  "(k, v) \<in> set (items f) \<longleftrightarrow> k \<in> Nat_Mapping.keys f \<and> Nat_Mapping.lookup f k = v"
  unfolding items_unfold by (auto simp add: in_keys_less_degree)

lemma sorted_map_fst_items [simp]:
  "sorted (map fst (items f))"
  unfolding items_unfold' by (rule sorted_filter) simp

lemma distinct_map_fst_items [simp]:
  "distinct (map fst (items f))"
  unfolding items_unfold' by (rule distinct_map_filter) simp

definition foldr :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b \<Rightarrow> 'b"
where
  "foldr f p = List.foldr (\<lambda>(k, v) c. f k v c) (items p)"

lemma foldr_zero [simp]:
  "foldr f 0 = id"
  unfolding foldr_def by (simp add: id_def)

lemma foldr_one [simp]:
  "foldr f 1 z = f 0 1 z"
  unfolding foldr_def by simp

lemma foldr_single [simp]:
  "foldr f (Nat_Mapping.single k v) z = (if v = 0 then z else f k v z)"
  unfolding foldr_def by simp

lemma foldr_cong [fundef_cong]:
  "\<lbrakk> m = m'; \<And>k r. k \<in> Nat_Mapping.keys m' \<Longrightarrow> f k (Nat_Mapping.lookup m' k) r = f' k (Nat_Mapping.lookup m' k) r \<rbrakk>
  \<Longrightarrow> foldr f m = foldr f' m'"
  unfolding foldr_def by (rule ext) (auto intro!: foldr_cong [OF refl])

lemma foldr_Dense:
  "foldr f (Dense xs) z = foldr_dense f xs z"
  unfolding foldr_def items_unfold
  apply transfer
  apply transfer
  apply (auto simp add: degree_nth foldr_dense_raw_conv_foldr foldr_filter
    enumerate_map_upt [symmetric] map_nth_default
    intro!: List.foldr_cong)
  done

lemma foldr_Sparse:
  "foldr f (Sparse xs) z = foldr_sparse f xs z"
  unfolding foldr_def by transfer
    (auto simp add: of_oalist_comparator_of_le_conv_the_value items_the_value)

lemmas foldr_code [code] =
  foldr_Dense
  foldr_Sparse

definition foldr_step :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (nat \<Rightarrow> nat option \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 
  (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b \<Rightarrow> 'b"
where "foldr_step f g p x = 
  (let (c, n') = foldr (\<lambda>n v (c, n'). (f n v (g n n' c), Some n)) p (x, None)
   in g 0 n' c)"

lemma foldr_step_cong [fundef_cong]:
  assumes *: "m = m'" "g = g'" (* FIXME: add sensible conditions for g *)
  and f: "\<And>k r. k \<in> Nat_Mapping.keys m' \<Longrightarrow> f k (Nat_Mapping.lookup m' k) r = f' k (Nat_Mapping.lookup m' k) r"
  shows "foldr_step f g m = foldr_step f' g' m'"
proof -
  have "Nat_Mapping_Recursive.foldr (\<lambda>n v (c, n'). (f n v (g' n n' c), Some n)) m' =
        Nat_Mapping_Recursive.foldr (\<lambda>n v (c, n'). (f' n v (g' n n' c), Some n)) m'"
    by(rule Nat_Mapping_Recursive.foldr_cong)(auto simp add: f)
  then show ?thesis
    by(simp add: split_beta fun_eq_iff foldr_step_def *)
qed

subsection \<open>Monotonic transformation on keys\<close>

lift_definition add_keys_pm :: "nat \<Rightarrow> (nat \<Rightarrow>\<^sub>0 'a :: zero) \<Rightarrow> (nat \<Rightarrow>\<^sub>0 'a)"
is "\<lambda>n f x. if x < n then 0 else f (x - n)"
proof -
  fix n :: nat and f :: "nat \<Rightarrow> 'a"
  let ?f = "\<lambda>x. if x < n then 0 else f (x - n)"
  assume *: "finite {x. f x \<noteq> 0}"
  have "{x. ?f x \<noteq> 0} \<subseteq> (\<lambda>x. x + n) ` {x. f x \<noteq> 0}"
    by(auto intro: rev_image_eqI)
  moreover have "finite \<dots>" using * by(rule finite_imageI)
  ultimately show "finite {x. ?f x \<noteq> 0}" by(rule finite_subset)
qed

lemma lookup_add_keys_pm [simp]:
  "Poly_Mapping.lookup (add_keys_pm n p) k = (if k < n then 0 else Poly_Mapping.lookup p (k - n))"
by transfer simp

lemma add_keys_pm_nth:
  "add_keys_pm n (Poly_Mapping.nth xs) = Poly_Mapping.nth (if xs = [] then [] else replicate n 0 @ xs)"
by transfer(auto simp add: nth_default_append)

lemma lookup_default_map_apfst_plus:
  "c.lookup_default comparator_of_le dflt (map (apfst (op + n)) xs) k =
  (if k < n then dflt else c.lookup_default comparator_of_le dflt xs (k - n :: nat))"
by(induction xs)(auto cong: comparison.case_cong simp add: comparator_of_le_def[abs_def])

lift_definition add_keys :: "nat \<Rightarrow> 'a :: zero nat_mapping \<Rightarrow> 'a nat_mapping"
is "add_keys_pm" .

lemma lookup_add_keys [simp]:
  "Nat_Mapping.lookup (add_keys n p) k = (if k < n then 0 else Nat_Mapping.lookup p (k - n))"
by transfer simp

lemma add_keys_nth:
  "add_keys n (Nat_Mapping.nth xs) = Nat_Mapping.nth (if xs = [] then [] else replicate n 0 @ xs)"
by transfer(simp add: add_keys_pm_nth)

lift_definition add_keys_dense :: "nat \<Rightarrow> 'a :: zero nat_mapping_dense \<Rightarrow> 'a nat_mapping_dense"
is "\<lambda>n xs. if xs = [] then [] else replicate n 0 @ xs"
by(simp)

lemma add_keys_Dense [code]: "add_keys n (Dense xs) = Dense (add_keys_dense n xs)"
by transfer(simp add: add_keys_nth)

lemma (in linordered_ab_semigroup_add) sorted_map_plus: "sorted xs \<Longrightarrow> sorted (map (plus x) xs)"
by(induction rule: sorted.induct)(auto simp add: sorted_Cons intro: add_left_mono)

lemma (in cancel_ab_semigroup_add) distinct_map_plus: "distinct xs \<Longrightarrow> distinct (map (plus x) xs)"
by(induction xs) auto

lift_definition add_keys_sparse :: "nat \<Rightarrow> 'a :: zero nat_mapping_sparse \<Rightarrow> 'a nat_mapping_sparse"
is "\<lambda>n. map (apfst (op + n))"
by simp(simp add: sorted_map_plus distinct_map_plus map_map[symmetric] image_image del: map_map)

lemma add_keys_mp_of_oalist: 
  "add_keys_pm n (of_oalist comparator_of_le xs) = of_oalist comparator_of_le (map (apfst (op + n)) xs)"
by transfer(simp add: fun_eq_iff lookup_default_map_apfst_plus)

lemma add_keys_Sparse [code]: "add_keys n (Sparse xs) = Sparse (add_keys_sparse n xs)"
by transfer(simp add: add_keys_mp_of_oalist)

hide_const (open) zip items foldr

end
