(* Author: Andreas Lochbihler, ETH Zurich
   Author: Florian Haftmann, TU Muenchen 
*)

section \<open>A recursive polynomial representation\<close>

theory Recursive
imports MPoly Recursive_Construct Nat_Mapping_Recursive
begin

text \<open>
  This is a variant of "~~/src/HOL/Library/Polynomial.thy" with the difference
  that the number of variables is not encoded in the type, i.e. @{text "'a poly_rec"}
  is the same type for any numbers of variables.
  Here is a preliminary version aiming at code generation at an early stage
  of bootstrapping a "Polynomial Package". This version will serve comparisons
  in efficiency and in conciseness with other versions.
  There are plans to replace this version with another version building upon
  @{text "'a nat_mapping"}. This would generalise the handling of dense versus
  sparse representations, etc. The difficulty with @{text "'a nat_mapping"} is, 
  that it is no bounded natural functor (BNF), thus cannot be implemented
  using @{text datatype} or even @{text datatype_new}.
\<close>

fun Rec_aux :: "nat \<Rightarrow> 'a :: semiring_1 poly_rec \<Rightarrow> 'a mpoly"
where
  "Rec_aux n (Coeff x) = monom 0 x"
| "Rec_aux n (Powers ps) =
   sum_nat_mapping (\<lambda>e p. Rec_aux (Suc n) p * monom (Poly_Mapping.single n e) 1) ps"

lemma Rec_aux_zero [simp]: "Rec_aux n 0 = 0"
by(simp add: zero_poly_rec_def)

definition Rec :: "'a :: semiring_1 poly_rec \<Rightarrow> 'a mpoly"
where "Rec = Rec_aux 0"

code_datatype Rec

definition rec_of :: "'a :: zero mpoly \<Rightarrow> 'a poly_rec"
where "rec_of = undefined"

lemma Rec_rec_of: "Rec (rec_of p) = p" sorry

lemma rec_of_Rec: "rec_of (Rec p) = p" oops
-- \<open>@{typ "_ poly_rec"} does not have unique representations, so this lemma is false.\<close>


subsection \<open>Normal form for this kind of polynomial\<close>

fun drop_leading_zeros :: "'a::zero raw_poly_rec list \<Rightarrow> 'a raw_poly_rec list"
where
  "drop_leading_zeros [] = []"
| "drop_leading_zeros (Coeff_raw c # ps) = (if c = 0 then drop_leading_zeros ps else Coeff_raw c # ps)"
| "drop_leading_zeros ps = ps"

definition drop_trailing_zeros :: "'a::zero raw_poly_rec list \<Rightarrow> 'a raw_poly_rec list"
where
  "drop_trailing_zeros = (rev \<circ> drop_leading_zeros \<circ> rev)"

function cut_trailing_zeros :: "'a::zero raw_poly_rec \<Rightarrow> 'a raw_poly_rec"
where
  "cut_trailing_zeros (Coeff_raw c) = Coeff_raw c"
| "cut_trailing_zeros (Powers_raw ps) = (let ps' = drop_trailing_zeros ps
    in Powers_raw (map cut_trailing_zeros ps'))"
by pat_completeness auto
termination sorry

(*this cuts only 1 level, so needs repetition in normalise*)
fun cut_empty_powers :: "'a::zero raw_poly_rec \<Rightarrow> 'a raw_poly_rec"
where
  "cut_empty_powers (Coeff_raw c) = Coeff_raw c"
| "cut_empty_powers (Powers_raw []) = Coeff_raw 0"
| "cut_empty_powers (Powers_raw [Coeff_raw c]) = Coeff_raw c"
(*"... this normalisation still allows to distinguish certain polynomials: 
  e.g. in (R[x])[y] the following two polynomials:
    Powers [Powers [Coeff 0, Coeff 2]] \<longrightarrow> no normalisation
      2.x = y^0  . (x^0  .0 + x^1 .2)
    Powers [Coeff 0, Powers [Coeff 3]] \<longrightarrow> Powers [Coeff 0 , Coeff 3]
      3.y = y^0  .0 + y^1  .(x^0  .3)   =          y^0  .0 + y^1  .3) *)
| "cut_empty_powers (Powers_raw ps) = Powers_raw (map cut_empty_powers ps)"

(*cut_trailing_zeros can make no_empty_powers false and vice versa, thus repetition;
  tests no_trailing_zeros, no_empty_powers as costly as cut_trailing_zeros, cut_empty_powers;
  thus tests with no_* at a minimum.*)
function normalise :: "'a::zero raw_poly_rec \<Rightarrow> 'a raw_poly_rec"
where
  "normalise p = 
    (let p' = cut_trailing_zeros p;
      p'' = cut_empty_powers p'
    in if no_trailing_zeros_raw_poly_rec p''
      then p'' 
      else normalise p'')"
by pat_completeness auto
termination sorry

subsection \<open>Fundamentals: @{const Groups.zero}, @{const Groups.one}\<close>

subsection \<open>Fold combinator for polynomials\<close>
subsection \<open>Degree\<close>

fun degree_rec :: "'a :: zero poly_rec \<Rightarrow> nat \<Rightarrow> nat"
where
  "degree_rec (Coeff x) n   = 0"
| "degree_rec (Powers ps) n =
  (if n = 0 then Max (insert 0 (Nat_Mapping.keys ps))
   else Nat_Mapping_Recursive.foldr (\<lambda>_ p r. max (degree_rec p (n - 1)) r) ps 0)"

fun is_zero_rec :: "'a :: zero poly_rec \<Rightarrow> bool"
where
  "is_zero_rec (Coeff x) \<longleftrightarrow> x = 0"
| "is_zero_rec (Powers ps) \<longleftrightarrow> (\<forall>p \<in> Nat_Mapping.range ps. is_zero_rec p)"

fun total_degree_rec :: "'a :: zero poly_rec \<Rightarrow> nat"
where
  "total_degree_rec (Coeff x) = 0"
| "total_degree_rec (Powers ps) =
   Nat_Mapping_Recursive.foldr (\<lambda>n p r. if is_zero_rec p then r else max (n + total_degree_rec p) r) ps 0"

subsection \<open>Monomials\<close>

fun shift_rec :: "'a poly_rec \<Rightarrow> nat \<Rightarrow> 'a :: zero poly_rec"
where
  "shift_rec (Coeff x) n = Coeff x"
| "shift_rec (Powers ps) 0 = Powers ps"
| "shift_rec (Powers ps) (Suc n) = Powers (Nat_Mapping.single' (implT ps) 0 (shift_rec (Powers ps) n))"

definition monom_rec :: "nat nat_mapping \<Rightarrow> 'a :: zero \<Rightarrow> 'a poly_rec"
where
  "monom_rec m c =
  (if c = 0 then Coeff 0 else
   case_prod shift_rec (Nat_Mapping_Recursive.foldr
     (\<lambda>v n (p, v'). (Powers (Nat_Mapping.single n (shift_rec p (v' - v))), v))
     m (Coeff c, 0)))"
 
definition monom_rec' :: "Nat_Mapping.impl \<Rightarrow> nat nat_mapping \<Rightarrow> 'a :: zero \<Rightarrow> 'a poly_rec"
where [simp, code del]: "monom_rec' _ = monom_rec"

lemma monom_rec'_code [code]:
  "monom_rec' impl m c =
  (if c = 0 then Coeff 0 else
   case_prod shift_rec (Nat_Mapping_Recursive.foldr 
     (\<lambda>v n (p, v'). (Powers (Nat_Mapping.single' impl n (shift_rec p (v' - v))), v))
     m (Coeff c, 0)))"
by(simp add: monom_rec_def)


subsection \<open>Addition and subtraction\<close>

instantiation poly_rec :: (monoid_add) monoid_add
begin

fun plus_poly_rec :: "'a poly_rec \<Rightarrow> 'a poly_rec \<Rightarrow> 'a poly_rec"
where
  "(Coeff x)   + (Coeff y)   = Coeff (x + y)"
| "(Powers ps) + (Powers qs) =
   Powers (Nat_Mapping_Recursive.zip plus_poly_rec ps qs)"
| "(Coeff x)   + (Powers qs) =
   Powers (Nat_Mapping_Recursive.zip plus_poly_rec (Nat_Mapping.single' (implT qs) 0 (Coeff x)) qs)"
| "(Powers ps) + (Coeff y)   =
   Powers (Nat_Mapping_Recursive.zip plus_poly_rec ps (Nat_Mapping.single' (implT ps) 0 (Coeff y)))"

instance
  sorry

end

declare plus_poly_rec.simps [simp del]

lemma plus_poly_rec_simps [simp, code]:
  "Coeff x   + Coeff y   = Coeff (x + y)"
  "Powers ps + Powers qs = Powers (ps + qs)"
  "Coeff x   + Powers qs = Powers (Nat_Mapping.single' (implT qs) 0 (Coeff x) + qs)"
  "Powers ps + Coeff y   = Powers (ps + Nat_Mapping.single' (implT ps) 0 (Coeff y))"
  using plus_poly_rec.simps unfolding Nat_Mapping_Recursive.zip_plus_plus by simp_all

lemma plus_rec_aux: "Rec_aux n (p + q) = Rec_aux n p + Rec_aux n q"
proof(induction p q arbitrary: n rule: plus_poly_rec.induct)
  case 1 thus ?case by(simp add: monom_add)
next
  case (2 ps qs)
  let ?f = "\<lambda>e p. Rec_aux (Suc n) p * monom (Poly_Mapping.single n e) 1"
  have "sum_nat_mapping ?f (ps + qs) = sum_nat_mapping ?f ps + sum_nat_mapping ?f qs"
    by(subst sum_nat_mapping_plus_distrib)(simp_all add: "2.IH" ring_distribs)
  thus ?case by simp
next
  case (3 x qs)
  let ?f = "\<lambda>e p. Rec_aux (Suc n) p * monom (Poly_Mapping.single n e) 1"
  have "sum_nat_mapping ?f (Nat_Mapping.single 0 (Coeff x) + qs) = monom 0 x + sum_nat_mapping ?f qs"
    using "3.IH" by(subst sum_nat_mapping_plus_distrib)(simp_all add: when_def mult_monom)
  then show ?case by simp
next
  case (4 ps y)
  let ?f = "\<lambda>e p. Rec_aux (Suc n) p * monom (Poly_Mapping.single n e) 1"
  have "sum_nat_mapping ?f (ps + Nat_Mapping.single 0 (Coeff y)) = sum_nat_mapping ?f ps + monom 0 y"
    using "4.IH" by(subst sum_nat_mapping_plus_distrib)(simp_all add: when_def mult_monom)
  then show ?case by simp
qed

lemma plus_rec [code]: "Rec p + Rec q = Rec (p + q)"
by(simp add: plus_rec_aux Rec_def)


subsection \<open>Multiplication by a coefficient an increment of a power\<close>

fun smult_rec :: "'a \<Rightarrow> 'a :: {zero, times} poly_rec \<Rightarrow> 'a poly_rec"
where
    "smult_rec s (Coeff c) = Coeff (s * c)"
  | "smult_rec s (Powers ps) = 
      (if s = 0 then Coeff 0 else Powers (Nat_Mapping.mapp (\<lambda>_. smult_rec s) ps))"

lemma smult_Rec_aux:
  "Rec_aux n (smult_rec s p) = smult s (Rec_aux n p)"
proof(induction p arbitrary: n)
  case Coeff thus ?case by simp
next
  case (Powers ps)
  have "sum_nat_mapping (\<lambda>e x. Rec_aux (Suc n) (smult_rec s x) * monom (Poly_Mapping.single n e) 1) ps =
        sum_nat_mapping (\<lambda>e x. smult s (Rec_aux (Suc n) x * monom (Poly_Mapping.single n e) 1)) ps"
    by(rule sum_nat_mapping_cong[OF refl])(simp add: Powers.IH mult_smult_left)
  then show ?case
    by(simp add: sum_nat_mapping_mapp sum_nat_mapping_right_distrib smult_conv_mult)
qed

lemma smult_rec [code]: "smult s (Rec p) = Rec (smult_rec s p)"
by(simp add: Rec_def smult_Rec_aux)

fun inc_power_rec :: "Nat_Mapping.impl \<Rightarrow> nat \<Rightarrow> 'a :: {zero} poly_rec \<Rightarrow> 'a poly_rec"
where
  "inc_power_rec impl 0 (Coeff a) = 
    (if a = 0 then Coeff 0 else Powers (Nat_Mapping.single' impl 1 (Coeff a)))"
| "inc_power_rec _ 0 (Powers ps) = Powers (add_keys 1 ps)"
| "inc_power_rec impl lev (Coeff a) = 
    (if a = 0 then Coeff 0 else 
      Powers (Nat_Mapping.single' impl 0 (inc_power_rec impl (lev - 1) (Coeff a))))"
| "inc_power_rec impl lev (Powers ps) = 
    Powers (Nat_Mapping.mapp (\<lambda>_. inc_power_rec impl (lev - 1)) ps)"

subsection \<open>Polynomial multiplication\<close>

(* might be shifted to some auxiliary stuff *)
fun funpow :: "nat \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a)"
where
  "funpow 0 f = (\<lambda>x. x)"
| "funpow (Suc n) f = (f o (funpow n f))"

(*ad "fun times" see mail Andreas Re: Nat_Mapping_Recursive.foldr Thu, 22 Jan 2015 18:18:14 +0100
  and c449292e2133 for termination *)
fun times :: "nat \<Rightarrow> 'a :: {times,monoid_add} poly_rec \<Rightarrow> 'a poly_rec \<Rightarrow> 'a poly_rec"
where
  "times _ (Coeff a) (Coeff b) = Coeff (a * b)"
| "times _ (Coeff a) (Powers qs) = Powers (Nat_Mapping.mapp (\<lambda>_. smult_rec a) qs)"
| "times _ (Powers ps) (Coeff b) = Powers (Nat_Mapping.mapp (\<lambda>_. smult_rec b) ps)"
| "times lev (Powers ps) q = Nat_Mapping_Recursive.foldr_step 
    (\<lambda>_ a p. (times (lev + 1) a q) + p)
    (\<lambda>n n' p. 
      let n'' = (case n' of None \<Rightarrow> 0 | Some n'' \<Rightarrow> n'')
      in funpow (n'' - n) (inc_power_rec (implT ps) lev) p) 
    ps (Coeff 0)"

instantiation poly_rec :: (semiring_0) semiring_0
begin

lift_definition times_poly_rec :: "'a poly_rec \<Rightarrow> 'a poly_rec \<Rightarrow> 'a poly_rec"
is "times 0" sorry

instance sorry

end

lemma times_rec [code]:
  "Rec p * Rec q = Rec (p * q)"
sorry

subsection \<open>Pseudo division --- preliminarily only univariate case\<close>
(* started cp from mult above:

instantiation poly_rec :: (ring_1_no_zero_divisors) ring_div
begin

lift_definition pdiv_poly_rec :: "'a poly_rec \<Rightarrow> 'a poly_rec \<Rightarrow> ('a * 'a poly_rec)"
is "times 0" sorry

instance sorry

end

lemma times_rec [code]:
  "Rec p * Rec q = Rec (p * q)"
sorry
*)

end