(* Author: Andreas Lochbihler, ETH Zurich *)

section \<open>Constructing the recursive representation for multivariate polynomials\<close>

theory Recursive_Construct
imports Nat_Mapping
begin



subsection \<open>Type of recursive polynomial representation\<close>

(*SEE MAIL On 03/27/2014 12:33 PM, Andreas Lochbihler Re: Rekursive Darstellung*)
datatype 'a raw_poly_rec = Coeff_raw 'a | Powers_raw "('a raw_poly_rec) list"

instantiation raw_poly_rec :: (zero) zero begin
definition "0 = Coeff_raw 0"
instance ..
end

(*SEE MAIL On  01/20/2014 09:29 AM, Andreas Lochbihler: Re: ?abstract? poly*)
fun no_trailing_zeros_raw_poly_rec :: "'a :: zero raw_poly_rec \<Rightarrow> bool"
where
  "no_trailing_zeros_raw_poly_rec (Coeff_raw _)   \<longleftrightarrow> True"
| "no_trailing_zeros_raw_poly_rec (Powers_raw ps) \<longleftrightarrow> 
   no_trailing_zeros ps \<and> (\<forall>p\<in>set ps. no_trailing_zeros_raw_poly_rec p)"


typedef (overloaded) 'a poly_rec = "{p :: 'a :: zero raw_poly_rec. no_trailing_zeros_raw_poly_rec p}"
proof
  show "(Coeff_raw 0) \<in> ?poly_rec" by simp
qed

setup_lifting type_definition_poly_rec

subsection \<open>Constructors\<close>

lift_definition Coeff :: "'a :: zero \<Rightarrow> 'a poly_rec" is "Coeff_raw :: 'a \<Rightarrow> _" by simp

instantiation poly_rec :: (zero) zero begin
definition "0 = Coeff 0"
instance ..
end

lemma Rep_poly_rec_eq_0_iff: "Rep_poly_rec x = 0 \<longleftrightarrow> x = 0"
unfolding zero_poly_rec_def by transfer(simp add: zero_raw_poly_rec_def)

instantiation poly_rec :: ("{zero, one}") one begin
definition "1 = Coeff 1"
instance ..
end

definition Powers :: "'a :: zero poly_rec nat_mapping \<Rightarrow> 'a poly_rec"
where "Powers ps = Abs_poly_rec (Powers_raw (map Rep_poly_rec (Rep_nat_mapping_dense (rep_dense ps))))"

code_datatype Coeff Powers

lemma Coeff_inject: "Coeff x = Coeff y \<longleftrightarrow> x = y"
by transfer simp

lemma no_trailing_zeros_Powers [simp]:
  "no_trailing_zeros (map Rep_poly_rec (Rep_nat_mapping_dense m))"
  using Rep_nat_mapping_dense[of m]
  by(clarsimp simp add: no_trailing_unfold last_map zero_raw_poly_rec_def zero_poly_rec_def Coeff_def Rep_poly_rec_inverse)

lemma no_trailing_zeros_raw_poly_rec_Rep_poly_rec [simp]:
  "no_trailing_zeros_raw_poly_rec (Rep_poly_rec x)"
  using Rep_poly_rec[of x] by simp

lemma Powers_rep_eq:
  "Rep_poly_rec (Powers ps) = Powers_raw (map Rep_poly_rec (Rep_nat_mapping_dense (rep_dense ps)))"
by(simp add: Powers_def Abs_poly_rec_inverse)

lemma inj_Rep_poly_rec [simp]: "inj Rep_poly_rec"
by(simp add: inj_on_def Rep_poly_rec_inject)

lemma Powers_inject: "Powers ps = Powers ps' \<longleftrightarrow> ps = ps'"
by(simp add: Powers_def Abs_poly_rec_inject Rep_nat_mapping_dense_inject rep_dense_inject)

lemma Coeff_neq_Powers: "Coeff x \<noteq> Powers ps"
by(simp add: Coeff_def Powers_def Abs_poly_rec_inject)

lemma poly_rec_induct[case_names Coeff Powers, induct type: poly_rec]:
  assumes Coeff: "\<And>x. P (Coeff x)"
  and Powers: "\<And>ps. (\<And>p. p \<in> Nat_Mapping.range ps \<Longrightarrow> P p) \<Longrightarrow> P (Powers ps)"
  shows "P p"
proof(cases p rule: Abs_poly_rec_cases)
  case (Abs_poly_rec p')
  hence "no_trailing_zeros_raw_poly_rec p'" by simp
  hence "P (Abs_poly_rec p')"
  proof(induction p' rule: no_trailing_zeros_raw_poly_rec.induct)
    case (1 x)
    from Coeff[of x] show ?case
      by(simp add: Coeff_def)
  next
    case (2 ps)
    let ?ps = "Dense (Abs_nat_mapping_dense (map Abs_poly_rec ps))"

    have *: "no_trailing_zeros (map Abs_poly_rec ps)" using "2.prems"
      by(clarsimp simp add: no_trailing_unfold last_map zero_poly_rec_def Coeff_def Abs_poly_rec_inject zero_raw_poly_rec_def)
    moreover
    have "P (Powers (Dense (Abs_nat_mapping_dense (map Abs_poly_rec ps))))"
    proof(rule Powers)
      fix p
      assume "p \<in> Nat_Mapping.range ?ps"
      hence "Rep_poly_rec p \<in> set ps" using * "2.prems"
        by(auto simp add: range_Dense range_dense_def Abs_nat_mapping_dense_inverse Abs_poly_rec_inverse)
      hence "P (Abs_poly_rec (Rep_poly_rec p))" using "2.prems"
        by(auto intro: "2.IH")
      thus "P p" by(simp add: Rep_poly_rec_inverse)
    qed
    moreover have "map (Rep_poly_rec \<circ> Abs_poly_rec) ps = map id ps" using "2.prems"
      by(intro map_cong[OF refl])(simp add: Abs_poly_rec_inverse)
    ultimately show ?case
      by(simp add: Powers_def rep_dense_Dense Abs_nat_mapping_dense_inverse)
  qed
  thus "P p" by(simp add: Abs_poly_rec)
qed

lemma poly_rec_exhaust:
  obtains (Coeff) y where "x = Coeff y"
   | (Powers) ps where "x = Powers ps"
by(induction x) auto

free_constructors case_poly_rec for Coeff | Powers
by(metis poly_rec_exhaust)(simp_all add: Coeff_inject Powers_inject Coeff_neq_Powers)

lemma [simp]:
  shows Powers_neq_0: "Powers ps \<noteq> 0"
  and "0_neq_Powers": "0 \<noteq> Powers ps"
  and Coeff_eq_0_iff: "Coeff x = 0 \<longleftrightarrow> x = 0"
  and "0_eq_Coeff_iff": "0 = Coeff x \<longleftrightarrow> x = 0"
by(simp_all add: zero_poly_rec_def)

lemma list_all2_dropWhileI:
  "\<lbrakk> list_all2 R xs ys; \<And>x y. \<lbrakk> (x, y) \<in> set (zip xs ys); R x y \<rbrakk> \<Longrightarrow> P x \<longleftrightarrow> Q y \<rbrakk>
  \<Longrightarrow> list_all2 R (dropWhile P xs) (dropWhile Q ys)"
by(induction xs ys rule: list_all2_induct) auto

lemma list_all2_strip_whileI:
  "\<lbrakk> list_all2 R xs ys; \<And>x y. \<lbrakk> (x, y) \<in> set (zip xs ys); R x y \<rbrakk> \<Longrightarrow> P x \<longleftrightarrow> Q y \<rbrakk>
  \<Longrightarrow> list_all2 R (strip_while P xs) (strip_while Q ys)"
by(auto simp add: strip_while_def zip_rev dest: list_all2_lengthD intro!: list_all2_dropWhileI)

lift_definition rec_poly_rec :: "('a :: zero \<Rightarrow> 'b) \<Rightarrow> (('a poly_rec \<times> 'b) list \<Rightarrow> 'b) \<Rightarrow> 'a poly_rec \<Rightarrow> 'b"
is "\<lambda>c f. rec_raw_poly_rec c (\<lambda>xs. f (strip_while (\<lambda>(p, x). p = 0) xs))"
proof -
  fix c :: "'a \<Rightarrow> 'b"
    and f1 :: "('a raw_poly_rec \<times> 'b) list \<Rightarrow> 'b" 
    and f2 :: "('a raw_poly_rec \<times> 'b) list \<Rightarrow> 'b"
    and p :: "'a raw_poly_rec"
  let ?R = "pred_prod no_trailing_zeros_raw_poly_rec top"
  let ?f1 = "\<lambda>xs. f1 (strip_while (\<lambda>(p, x). p = 0) xs)"
  let ?f2 = "\<lambda>xs. f2 (strip_while (\<lambda>(p, x). p = 0) xs)"
  assume f1_f2: "\<And>xs. list_all ?R xs \<Longrightarrow> f1 xs = f2 xs"
  assume "no_trailing_zeros_raw_poly_rec p"
  then show "rec_raw_poly_rec c ?f1 p = rec_raw_poly_rec c ?f2 p"
  proof (induction p)
    case (Powers_raw xs)
    find_theorems "Powers_raw xs"
    have "Powers_raw xs \<in> set xs"
      sorry
    then show ?case using Powers_raw.prems
      by (rule Powers_raw.IH)
  qed simp
qed

lemma rec_poly_rec_Coeff: "rec_poly_rec c p (Coeff x) = c x"
by transfer simp

lemma rec_poly_rec_Powers:
  "rec_poly_rec c p (Powers ps) =
  p (map (\<lambda>r. (r, rec_poly_rec c p r)) (Rep_nat_mapping_dense (rep_dense ps)))"
by(simp add: rec_poly_rec_def Powers_rep_eq strip_while_map o_def Rep_poly_rec_inverse Rep_poly_rec_eq_0_iff strip_while_Rep_nat_mapping_dense')

lifting_update poly_rec.lifting
lifting_forget poly_rec.lifting


subsection \<open>Size function on @{typ "_ poly_rec"}\<close>

inductive_set wf_size_rel :: "('a :: zero poly_rec \<times> 'a poly_rec) set"
where
  "p \<in> Nat_Mapping.range ps \<Longrightarrow> (p, Powers ps) \<in> wf_size_rel"
| "p \<noteq> 0 \<Longrightarrow> (0, p) \<in> wf_size_rel"

inductive_simps wf_size_rel_simps [simp]:
  "(p, Coeff x)   \<in> wf_size_rel"
  "(p, Powers ps) \<in> wf_size_rel"
  "(p, 0)         \<in> wf_size_rel"


lemma wf_wf_size_rel [simp]: "wf wf_size_rel"
unfolding wf_def
proof(intro strip)
  fix P :: "'a poly_rec \<Rightarrow> bool" and p
  assume IH [rule_format]: "\<forall>x. (\<forall>y. (y, x) \<in> wf_size_rel \<longrightarrow> P y) \<longrightarrow> P x"
  show "P p" by(induction p)(auto 4 4 intro: IH)
qed

context fixes f :: "'a :: zero \<Rightarrow> nat" begin

function poly_rec_size :: "'a poly_rec \<Rightarrow> nat"
where
  "poly_rec_size (Coeff x) = f x + 1"
| "poly_rec_size (Powers ps) = 1 + nat_mapping_size poly_rec_size ps"
by pat_completeness simp_all
termination by(relation "wf_size_rel") simp_all

end

instantiation poly_rec :: (zero) size begin
definition "size = poly_rec_size (\<lambda>_. 0)"
instance ..
end

lemma poly_rec_size_simps [simp, code]:
  shows size_Coeff: "size (Coeff x) = 1"
  and size_Powers: "size (Powers ps) = 1 + (nat_mapping_size size ps)"
  by (simp_all add: size_poly_rec_def)

lemma poly_rec_size_0 [simp]:
  "size (0 :: _ poly_rec) = 1"
  by (simp add: zero_poly_rec_def)

lemma poly_rec_size_1 [simp]:
  "size (1 :: _ poly_rec) = 1"
  by (simp add: one_poly_rec_def)

lemma poly_rec_Coeff_when [simp]:
  "size (Coeff x when k = 0) = 1"
  by (cases "k = 0") simp_all

end