(* Author: Wolfgang Schreiner, RISC Linz
   Author: Walther Neuper, TU Graz
*)

theory Poly_Demo 
imports Complex_Main MPoly_Impl
begin

section \<open>Joint work Computer Algebra --- Isabelle\<close>
text \<open>see http://www.ist.tugraz.at/projects/isac/publ/poly-demo.pdf\<close>
 
section \<open>Towards a Polynomial Package in Isabelle\<close>
text \<open>see http://www.ist.tugraz.at/projects/isac/publ/poly-demo.pdf\<close>

section \<open>Demo: the proof-of-concept prototype\<close>
text \<open>Legend: --> is <Ctrl> + mouse-click: acts as a hyperlink to definition\<close>

subsection \<open>Isabelle/jEdit: towards a prover IDE\<close>
text \<open>
  # parallel evaluation of theories: 
      edit MPoly.thy ... Distributive || Recursive
  # formal source text with rich formal markup: see below!
  # asynchronous proof processing: TODO during presentation
  # link to definitions: TODO during presentation
  # Documentation - sidebar ...\<close>

subsection \<open>An example algorithm and a proof\<close>

text \<open>An algorithm featured by the prototype already\<close>
definition double :: "'a::monoid_add mpoly \<Rightarrow> 'a mpoly"
where "double p = p + p"

text \<open>How does above "+" relate to "finite {x. f x \<noteq> 0}"? 
  "+" leads to syntax definition -- @{class monoid_add},
  relation defining semantics is plus_poly_mapping in Poly_Mapping.\<close>

text \<open>Algorithms can be executed within Isabelle/HOL:\<close>
value
  (* 0 \<Rightarrow>\<^sub>0 1: x\<^sup>1, 2 \<Rightarrow>\<^sub>0 4: z\<^sup>4, ...                    2 * x\<^sup>1  +  3 * z\<^sup>4 *) 
  "let p = monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 0 1) 2 +
           monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 2 4) (3::int)
  in (double p)"                              (*... 4 * x\<^sup>1  +  6 * z\<^sup>4 *)

text \<open>Given a definition, we prove a property of the algorithm:\<close>
definition primitive :: "'a::{ring_div, GCD.Gcd} mpoly \<Rightarrow> bool"
where "primitive p \<longleftrightarrow> Gcd (coeffs p) = 1"

lemma coeffs_plus_same: "coeffs (p + p) = 
  (\<lambda>c :: _ :: linordered_ab_group_add. c + c) ` coeffs p"
apply transfer
apply transfer
apply auto
done

lemma Gcd_image_times_int: 
  assumes "finite (A :: int set)"
  shows "Gcd (op * k ` A) = \<bar>k\<bar> * Gcd A"
proof -
  from assms obtain xs where "A = set xs" by(auto dest: finite_list)
  moreover
  have "\<And>c. fold (gcd \<circ> op * k) xs \<bar>k * c\<bar> = \<bar>k\<bar> * fold gcd xs \<bar>c\<bar>"
  proof(induction xs)
    case Nil thus ?case by(simp add: linordered_idom_class.abs_mult)
  next
    case (Cons x xs)
    from Cons.IH[of "gcd x c"]
    show ?case
      by(simp add: gcd_mult_distrib_int[symmetric])(simp add: linordered_idom_class.abs_mult)
  qed
  note this[of 0]
  ultimately show ?thesis 
    by(simp add: Gcd_set_int fold_map set_map[symmetric] del: set_map)
qed

lemma double_not_primitive:
  fixes p q :: "int mpoly"
  assumes "q = double p"
  shows "\<not> primitive q"
proof
  assume "primitive q"
  then have *: "Gcd (coeffs q) = 1" by (simp add: primitive_def)
  from assms have "coeffs q = coeffs (double p)" by simp
  then have "coeffs q = coeffs (p + p)" sorry
  then have "coeffs q = (\<lambda>c. c + c) ` coeffs p" sorry
  then have "Gcd (coeffs q) = 2 * Gcd (coeffs p)" 
    by (simp add: Gcd_image_times_int)
  then have "Gcd (coeffs q) \<noteq> 1" sorry
  with * show False by simp
qed (* ... Sledgehammer resolves the minor missing steps! *)

text \<open>Use the findings of Sledgehammer for a more concise proof
  and first add the required theorems to the simplifier:\<close>
declare zmult_eq_1_iff [simp add]
declare primitive_def [simp add]
declare double_def [simp add]
declare coeffs_plus_same [simp add]
declare Gcd_image_times_int [simp add]

lemma double_not_primitive2:
  fixes p q :: "int mpoly"
  assumes "q = double p"
  shows "\<not> primitive q"
proof
  assume "primitive q"
  moreover from this have "Gcd (coeffs q) = 1" by auto
  moreover from this have "coeffs (p + p) = (\<lambda>c. c + c) ` coeffs p" 
    by (metis coeffs_plus_same) 
  moreover from this assms have "Gcd (coeffs q) = 2 * Gcd(coeffs p)" by auto
  ultimately show False by auto
qed

lemma double_not_primitive3:
  fixes p q :: "int mpoly"
  assumes "q = double p"
  shows "\<not> primitive q"
proof
  assume "primitive q"
  moreover from this have "Gcd (coeffs q) = 1" by auto
  moreover from assms coeffs_plus_same[of "p"] have 
    "Gcd (coeffs q) = 2 * Gcd(coeffs p)" by auto
  ultimately show False by auto
qed

text \<open>Generalisation of double_not_primitive to "'a mpoly" 
  would require a more specific instatiation of Gcd.\<close>

subsection \<open>Automated code generation\<close>
text \<open>For executing code we need executable polynomial representations 
  -- the following are preliminary representations:\<close>
definition p011 :: "int mpoly" --\<open>distributive sparse\<close>
where "p011 = monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 0 1) 2 +
              monom' (DISTR RLEX) (Nat_Mapping.single' SPARSE 2 4) 3"

definition p012 :: "int mpoly" --\<open>distributive dense\<close>
where "p012 = monom' (DISTR RLEX) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (DISTR RLEX) (Nat_Mapping.single' DENSE 2 4) 3"

definition p013 :: "int mpoly" --\<open>recursive dense\<close>
where "p013 = monom' (REC DENSE) (Nat_Mapping.single' DENSE 0 1) 2 +
              monom' (REC DENSE) (Nat_Mapping.single' DENSE 2 4) 3"

definition p014 :: "complex mpoly" --\<open>recursive sparse\<close>
where "p014 = monom' (REC SPARSE) (Nat_Mapping.single' DENSE 0 1) (Complex 2 4) +
              monom' (REC SPARSE) (Nat_Mapping.single' DENSE 2 4) (Complex 3 5)"

lemma "p011 = p012" "p012 = p013" "p011 = p013"
by(simp_all add: p011_def p012_def p013_def)
-- \<open>we have logical equivalence for different representations of 
      polynomial @{text "2 * x\<^sup>1  +  3 * z\<^sup>4"} ...\<close> 

text \<open>We make things simple and instantiate "double" for integers:\<close>
definition double_int :: "int mpoly \<Rightarrow> int mpoly" --\<open>integers\<close>
where "double_int p = p + p"
value "p013"
value "double_int p013"
export_code double_int p013 in SML module_name Double_Int
ML \<open>\<close>
text \<open>open Double_Int\<close>
text \<open>double_int p013\<close>
ML \<open>\<close>

text \<open>Code from the abstract algorithm (p011, p014 included for executing)\<close>
export_code double p011 p014 in SML module_name Double
ML \<open> \<close>
ML \<open> \<close>
text \<open>open Double\<close>
text \<open>double\<close>
text \<open>p011\<close>
(* first copy this into the signature above:
  val one_int : int one
  val equal_int : int equal
  val ring_int : int ring
*)
text \<open>double (one_int, equal_int, ring_int) p011\<close>

(* first copy this into the signature above:
  val one_complex : complex one
  val equal_complex : complex equal
  val ring_complex : complex ring
*)
text \<open>double (one_complex, equal_complex, ring_complex) p014\<close>
ML \<open>\<close>

text \<open>Instantiate double for rational numbers:\<close>
definition double_rat :: "rat mpoly \<Rightarrow> rat mpoly" --\<open>rational numbers\<close>
where "double_rat p = p + p"
export_code double_rat in SML module_name Double_Rat

text \<open>Instantiate double for complex numbers:\<close>
definition double_complex :: "complex mpoly \<Rightarrow> complex mpoly" --\<open>complex numbers\<close>
where "double_complex p = p + p"
export_code double_complex p014 in SML module_name Double_Complex
ML \<open>\<close>
text \<open>open Double_Complex\<close>
text \<open>double_complex p014\<close>
ML \<open>\<close>

subsection \<open>Isabelle's mechanism ``typedef''\<close>
text \<open>
  typedef @{typ "'a mpoly"} --> poly_mapping --> finite,
  typedef: prove types not empty, enforces invariants,
  morphisms can be named (mapping_of, MPoly),
  helpful theorems generated from setup_lifting.\<close>

subsection \<open>Isabelle's mechanism ``instantiation''\<close>
text \<open>
  @{typ "'a::zero"} --> Groups.thy: class semigroup_add, ...
  @{text "lift_definition plus_mpoly"} in @{theory MPoly}
  @{text "plus_poly_mapping"} in @{theory Poly_Mapping}\<close>

section\<open>Conclusions\<close>
text \<open>
  Joint development lead to a proof-of-concept prototype,\\
  which successfully demonstrates:
    (1.) an abstract polynomial for proofs
    (2.) executable representations of polynomials
    (3.) where (1.) and (2.) are logically equivalent
    (4.) algorithms can be made efficient by code generation,\\
        where proofs from (1.) transfer to the code.
  which raises further questions:
    \<bullet> principal design for a polynomial package
    \<bullet> specific requirements w.r.t. Isabelle's BNF package, etc
  which makes promises for a joint project on a polynomial package:
    \<bullet> integrate design into Isabelle
    \<bullet> implement proof tools, parsers, etc
    \<bullet> mechanise parts of Franz Winkler's book "Polynomial Algorithms"/\<close>

end
